{$F+}
Unit
	Arkanoid; { Polit Halcyon Arkanoid application }

Interface

Uses
	Kernel, Types, D, Ctl, Gfx;

Const
	CaretWidth=60;
	BricksX=14;
	BricksY=9;
	BrickX=30;
	BrickY=15;

Type
	TArkanoidData=Record
		CanField, TimTicker, LabStatus: PControl;
		CaretPos, CaretMax: Integer;
		BX, BY, BDX, BDY: Integer;
		BColor: {TColor}Integer;
		Brick: Array[0..BricksX, 0..BricksY] Of Boolean;
		BricksLeft: Integer;
		State: (asNoGame, asGame, asBallLost, asPause);
	End;

Implementation

Procedure InitGame(Me:PProg);
Var
	X, Y: Integer;
Begin
	With Me^, ClientRect, TArkanoidData(Data^) Do
	Begin
		CaretMax:=X2-X1-6-CaretWidth;
		CaretPos:=CaretMax Div 2;
		BricksLeft:=0;
		For Y:=0 To BricksY Do
			For X:=0 To BricksX Do
				If Random(3)=1 Then
				Begin
					Brick[X, Y]:=True;
					Inc(BricksLeft)
				End
				Else Brick[X, Y]:=False;
		BX:=CaretPos+CaretWidth Div 2;
		BY:=260;
		BDX:=1;
		BDY:=-1;
		BColor:=200;
	End;
End;



Procedure OnPaint(Me:PProg);
Var
	X, Y: Integer;
Begin
	With Me^, ClientRect, TArkanoidData(Data^) Do
	Begin
		{InitGame(Me);}
		Ctl.Canvas_Paint(CanField, 3, 3, X2-X1-3, Y2-Y1-24);
		With CanField^.Place Do
		Begin
			Gfx.Bar(X1, Y1, X2, Y2, clBlack);
			If State<>asNoGame Then
			Begin
				Gfx.Bar(X1+CaretPos, Y2-9, X1+CaretPos+CaretWidth-1,
					Y2-1, clWhite);
				For Y:=0 To BricksY Do
					For X:=0 To BricksX Do
						If Brick[X, Y] Then
							Gfx.Bar(X1+1+X*BrickX, Y1+1+Y*BrickY,
								X1+1+(X+1)*BrickX-2, y1+1+(Y+1)*BrickY-2, clBlue);

				Gfx.Bar(X1+BX-2, Y1+BY-1, X1+BX+2, Y1+BY+1, BColor);
				Gfx.Line(X1+BX-1, Y1+BY-2, X1+BX+1, Y1+BY-2, BColor);
				Gfx.Line(X1+BX-1, Y1+BY+2, X1+BX+1, Y1+BY+2, BColor)
			End
		End;
		Ctl.Label_Paint(LabStatus, 3, Y2-Y1-21, X2-X1-3)
	End
End;

Procedure StartGame(Me:PProg);
Begin
	With TArkanoidData(me^.Data^) Do
	Begin
		State:=asGame;
		InitGame(Me);
		Kernel.PaintWindow(Me);
		Ctl.Label_SetCaption(LabStatus, 'i: '+D.St(BricksLeft))
	End
End;

Procedure EndGame(Me:PProg);
Begin
	With TArkanoidData(me^.Data^) Do
	Begin

	End
End;

Procedure MoveCaret(Me:PProg; Delta:Integer);
Begin
	With Me^, TArkanoidData(Data^), CanField^.Place Do
	Begin
		Kernel.BeginPaintCoord(X1, Y2-9, X2, Y2-1);
		Gfx.Bar(X1+CaretPos, Y2-9, X1+CaretPos+CaretWidth-1,
			Y2-1, clBlack);
		Inc(CaretPos, Delta);
		D.ToRange(CaretPos, 1, CaretMax);
		Gfx.Bar(X1+CaretPos, Y2-9, X1+CaretPos+CaretWidth-1,
			Y2-1, clWhite);
		Kernel.EndPaint
	End;
End;

Procedure OnTimeTicker(Me:PProg; Control:PControl; Times:Integer);
Var
	I, TX, TY: Integer;
	TX1, TY1, TX2, TY2: Integer;
	H, V: Integer;
	TempColor: TColor;
Begin
	With Me^, TArkanoidData(Data^) Do
	Begin
		If Not CanField^.CanDraw Then
		Begin
			Ctl.Label_SetCaption(LabStatus, '㧠');
			State:=asPause
		End;

		Case State Of
			asGame: Begin
					{ move caret }
				If (D.KeyStatus And ksLeftShift)<>0 Then MoveCaret(Me, -Times*10);
				If (D.KeyStatus And ksRightShift)<>0 Then MoveCaret(Me, Times*10);
					{ move ball }
				With CanField^.Place Do
					For I:=1 To 7 Do
					Begin
						Kernel.BeginPaintCoord(X1+BX-3, Y1+BY-3, X1+BX+3, Y1+BY+3);
						Gfx.Bar(X1+BX-2, Y1+BY-1, X1+BX+2, Y1+BY+1, clBlack);
						Gfx.Line(X1+BX-1, Y1+BY-2, X1+BX+1, Y1+BY-2, clBlack);
						Gfx.Line(X1+BX-1, Y1+BY+2, X1+BX+1, Y1+BY+2, clBlack);
						Kernel.EndPaint;
						Inc(BX, BDX);
						Inc(BY, BDY);
						If (BX=2) Or (BX=X2-X1-2) Then BDX:=-BDX;
						If (BY=2) Or (BY=Y2-Y1-2) Then BDY:=-BDY;
						Kernel.BeginPaintCoord(X1+BX+BDX*3, Y1+BY+BDY*3,
							X1+BX+BDX*3, Y1+BY+BDY*3);
						TempColor:=Gfx.GetPixel(X1+BX+BDX*3, Y1+BY+BDY*3);
						Kernel.EndPaint;
						If TempColor=clBlue Then
						Begin
							TX:=(BX+BDX*3-1) Div BrickX;
							TY:=(BY+BDY*3-1) Div BrickY;
							{Log.BugMsg('Arkanoid', 'OnTimeTicker', 'Deleting brick: X='+
								D.St(TX)+', Y='+D.St(TY));}
							Brick[TX, TY]:=False;
							Dec(BricksLeft);
							If BricksLeft<=10 Then
							Begin
								Kernel.EndPaint;
								State:=asNoGame;
								Ctl.Label_SetCaption(LabStatus, ' ࠫ! i - i...');
								Kernel.PaintWindow(Me);
								Break
							End;
							Ctl.Label_SetCaption(LabStatus, 'i: '+D.St(BricksLeft));
							TX1:=1+TX*BrickX;
							TY1:=1+TY*BrickY;
							TX2:=1+(TX+1)*BrickX-2;
							TY2:=1+(TY+1)*BrickY-2;
							Kernel.BeginPaintCoord(X1+BX+BDX*3-1, Y1+BY+BDY*3-1,
								X1+BX+BDX*3+1, Y1+BY+BDY*3+1);
							If Gfx.GetPixel(X1+BX+BDX*3-1, Y1+BY+BDY*3)=clBlue
								Then H:=1
								Else H:=0;
							If Gfx.GetPixel(X1+BX+BDX*3+1, Y1+BY+BDY*3)=clBlue
								Then Inc(H);
							If Gfx.GetPixel(X1+BX+BDX*3, Y1+BY+BDY*3-1)=clBlue
								Then V:=1
								Else V:=0;
							If Gfx.GetPixel(X1+BX+BDX*3, Y1+BY+BDY*3+1)=clBlue
								Then Inc(V);
							Kernel.EndPaint;
							Kernel.BeginPaintCoord(X1+TX1, Y1+TY1, X1+TX2, Y1+TY2);
							Gfx.Bar(X1+TX1, Y1+TY1, X1+TX2, Y1+TY2, clBlack);
							Kernel.EndPaint;
							If H=2 Then BDY:=-BDY;
							If V=2 Then BDX:=-BDX;
							If (H=1) And (V=1) Then
							Begin
								BDY:=-BDY;
								BDX:=-BDX
							End
						End;

						If BX=2 Then BDX:=1;
						If BX=X2-X1-2 Then BDX:=-1;
						If BY=2 Then BDY:=1;
						If BY=Y2-Y1-2 Then BDY:=-1;

						If BY=Y2-Y1-12 Then
							If D.InRange(BX, CaretPos, CaretPos+CaretWidth)
								Then BDY:=-1
								Else Begin
									State:=asNoGame;
									Ctl.Label_SetCaption(LabStatus, ' ணࠫ :(. i - i...');
									Kernel.PaintWindow(Me);
									Break
								End;

						D.Cycle(Integer(BColor), 200, 223);
						Kernel.BeginPaintCoord(X1+BX-2, Y1+BY-2, X1+BX+2, Y1+BY+2);
						Gfx.Bar(X1+BX-2, Y1+BY-1, X1+BX+2, Y1+BY+1, BColor);
						Gfx.Line(X1+BX-1, Y1+BY-2, X1+BX+1, Y1+BY-2, BColor);
						Gfx.Line(X1+BX-1, Y1+BY+2, X1+BX+1, Y1+BY+2, BColor);
						Kernel.EndPaint
					End

			End;

			(*asNoGame: Begin
				If (D.KeyStatus And $0f)<>0 Then { start game }
					StartGame(Me)
			End;*)
		End { case }
	End
End;

Function OnKey(Me:PProg; C1,C2:Char): Boolean;
Begin
	OnKey:=True;
	With TArkanoidData(Me^.Data^) Do
	Begin
		Case State Of
			asGame: If C1='p' Then
			Begin
				Ctl.Label_SetCaption(LabStatus, '㧠');
				State:=asPause
			End;
			asNoGame: StartGame(Me);
			asPause: Begin
				Ctl.Label_SetCaption(LabStatus, 'i: '+D.St(BricksLeft));
				State:=asGame
			End;
		End
	End
End;



Procedure OnClose(Me:PProg);
Begin
	FreeMem(Me^.Data, SizeOf(TArkanoidData))
End;

Procedure OnInit(Me:PProg);
Begin
	Me^.OnPaint:=OnPaint;
	Me^.OnKey:=OnKey;
	Me^.OnClose:=OnClose;
	With Me^ Do
	Begin
		D.AssignPStr('ઠ v0.1 (㢠 - i  ࠢ Shift)', WinTitle);
		MinX:=457;
		MinY:=322;
		GetMem(Data, SizeOf(TArkanoidData));
		Kernel.InitWindow(Me, 'BlueDlg', [wbClose]);
		Kernel.DrawWindowAni(Me);

		With TArkanoidData(Data^) Do
		Begin
			CanField:=Ctl.Canvas_Create(Me, False);
			TimTicker:=Ctl.TimerCtl_Create(Me, 1, True, OnTimeTicker);
			LabStatus:=Ctl.Label_Create(Me, 'i - i...',
				Nil, lsBeveled, alCenter);
			State:=asNoGame;
			{InitGame(Me);}
		End
	End
End;



Begin
	Kernel.RegBuiltIn('Arkanoid', ' ઠ v0.1', 'pic\i_arka.bmp', OnInit)
End.
