{$F+}
Unit
	Ctl3; { Polit Halcyon controls (interface) third unit }

Interface

Uses
	D, Log, Ini, Kernel, Types, Gfx, Font, Popup;

	Procedure Init;
	Procedure Done;

	Procedure HTML_Paint(Me:PControl; VX1,VY1,VX2,VY2:Integer);
	Procedure HTML_Draw(Me:PControl);
	Procedure HTML_LoadFile(Me:PControl; FName:String);
	Function HTML_Create(Prog:PProg): PControl;

Implementation

Uses
	Ctl;



{******* HTML *******}

Type
	TChunkKind=(chTag, chText);
	String18=String[18];
	PParam=^TParam;
	TParam=Record
		Name: String18;
		Value: PStr;
		Next: PParam;
	End;
	PChunk=^TChunk;
	TChunk=Record
		Next: PChunk;
		Case Kind: TChunkKind Of
			chTag: (
				Tag: String18;
				Opening: Boolean;
				FirstParam: PParam;
				SoftSpaces: Set Of (ssLeading, ssTailing);
			);
			chText: (Text: PStr);
	End;

	TPageItemKind=(piText, piLink, piImage, piRect, piBar,
		piPlaceholder, piTableStart);
	TTextAttr=Record
		Size, Color, Face: Integer;
		Attr: TVFontAttr;
		Align: TAlign;
	End;
	PPageItem=^TPageItem;
	TPageItem=Record
		Kind: TPageItemKind;
		Place: TRect;
		Text: PStr; { text, link text, image path }
		Attr: TTextAttr;
		Href: PStr; { link target, image link target }
		Image: Pointer;
		Color1, Color2: TColor;
		Prev, Next: PPageItem;
	End;

	TDocProperties=Record
		BGColor: Integer;
		TextColor: Integer;
		LinkColor: Integer;
		Title: String;
	End;
	TBGKind=(bgNone, bgColor, bgImage);

	THTML=Record
		LastWidth: Integer;
		HTMLScroll: PControl;
		MaxY: Integer;
	End;

Var
	FirstSrcLine, LastSrcLine: PPtrList;
	FirstChunk, LastChunk: PChunk;
	FirstPageItem, LastPageItem, LastPlaceholderItem: PPageItem;
	Area: Array[1..40] Of TRect;
	AreaNum: Integer;
	DocProps: TDocProperties;
	OrigMem: LongInt;
	MyFont: Pointer;
	HTMLMaxY, HTMLFirstY: Integer;



{$I Htm_Load.Inc}

{$I Htm_PreP.Inc}

{$I Htm_Pars.Inc}

{$I Htm_Draw.Inc}



Procedure HTML_IntDraw(Me:PControl; UpdateScroll:Boolean);
Begin
	If ((Not Me^.CanDraw) And (Not Kernel.DontBother)) Or Not Me^.Visible Then
		Exit;
	With Me^, THTML(Data^), Place Do
	Begin
		If LastWidth<>X2-X1+1 Then
		Begin
			If FirstPageItem<>Nil Then FreePageItems;
			ParseChunks(Me);

			LastWidth:=X2-X1+1
		End;

		Gfx.Rectangle(X1, Y1, X2, Y2, Kernel.ClrStatic);
		Gfx.Rectangle(X1+1, Y1+1, X2-1, Y2-1, Kernel.ClrWinBG);
		Gfx.Bar(X1+2, Y1+2, X2-2, Y2-2, DocProps.BGColor);

			{ call actual draw method }
		If HTMLMaxY<(Y2-Y1-5)
			Then HTMLFirstY:=0
			Else HTMLFirstY:=D.Min(HTMLFirstY, HTMLMaxY-(Y2-Y1-5));
		DrawPage(Me);

		If UpdateScroll Then
		Begin
			TControl(HTMLScroll^).Enabled:=Enabled;
			Ctl.Scroll_SetParams(HTMLScroll, (HTMLMaxY+9) Div 10,
				(Y2-Y1+9) Div 10, (HTMLFirstY+9) Div 10+1);
			If (HTMLScroll^.CanDraw Or Kernel.DontBother) And Visible Then
				Ctl.Scroll_PaintDirect(HTMLScroll, X2+2, Y1, X2+16, Y2)
		End
	End
End;

Procedure HTML_Draw(Me:PControl);
Begin
	With Me^ Do
		If Me^.CanDraw Or Kernel.DontBother Then
		Begin
			Kernel.BeginPaint(Place);
			If Visible
				Then HTML_IntDraw(Me, True)
				Else Kernel.HideCtl(Me);
			Kernel.EndPaint
		End
End;


Procedure HTML_Paint(Me:PControl; VX1,VY1,VX2,VY2:Integer);
Begin
	Log.BugMsg('HTML', 'HTML_Paint', 'Entering...');
	With Me^, Owner^.ClientRect Do
		D.AssignRect(X1+VX1, Y1+VY1, X1+VX2-16, Y1+VY2, Me^.Place);
	HTML_IntDraw(Me, True);
	Log.BugMsg('HTML', 'HTML_Paint', 'Done!');
End;

Procedure HTML_TryAutoRedraw(Me:PControl);
Begin
	If Me^.AutoRedraw And Not Kernel.IgnoreAutoRedraw Then
		If (Me^.CanDraw Or Kernel.DontBother) And Me^.Visible Then
		Begin
			Kernel.BeginPaint(Me^.Place);
			HTML_IntDraw(Me, True);
			Kernel.EndPaint
		End
End;

Procedure HTML_OnRefocus(Me:PControl; Focused:Boolean);
Begin
	Inc(Me^.Place.X2, 16);
	Kernel.CtlRefocusDummy(Me, Focused);
	Dec(Me^.Place.X2, 16)
End;





Procedure HTML_LoadFile(Me:PControl; FName:String);
Begin
	With Me^, THTML(Data^) Do
	Begin
		If FirstSrcLine<>Nil Then FreeSrcLines;
		If FirstChunk<>Nil Then FreeChunks;
		If FirstPageItem<>Nil Then FreePageItems;
		LastWidth:=0;
		HTMLFirstY:=0;
		LastPlaceholderItem:=Nil;

		If D.FileExist(FName) Then
		Begin
			LoadHTML(FName);
			BreakToChunks
		End
		Else
			Log.ErrMsg('HTML', 'HTML_LoadFile', 'File not found: '+FName);
		HTML_TryAutoRedraw(Me)
	End
End;

Procedure HTML_OnScrollUpdate(Me:PProg; Control:PControl);
Begin
	With THTML(Control^.Data^) Do
		HTMLFirstY:=(Ctl.Scroll_GetPos(HTMLScroll)-1)*10;
	HTML_IntDraw(Control, False)
End;

Function HTML_OnKey(Me:PControl; Key, ExtKey:Char):Boolean;
Var
	Recognized:Boolean;
Begin
	Recognized:=False;
	With Me^, Place, THTML(Data^) Do
	Begin
		If (Key=#0) And (ExtKey=#72) Then
		Begin { Up arrow }
			Dec(HTMLFirstY, 10);
			D.NotLess(HTMLFirstY, 0);
			HTML_IntDraw(Me, True);
			Recognized:=True
		End;
		If (Key=#0) And (ExtKey=#80) Then
		Begin { Down arrow }
			Inc(HTMLFirstY, 10);
			D.NotMore(HTMLFirstY, HTMLMaxY-(Y2-Y1+1));
			HTML_IntDraw(Me, True);
			Recognized:=True
		End;
		If (Key=#0) And (ExtKey=#73) Then
		Begin { PgUp }
			Dec(HTMLFirstY, Y2-Y1-30);
			D.NotLess(HTMLFirstY, 0);
			HTML_IntDraw(Me, True);
			Recognized:=True
		End;
		If (Key=#0) And (ExtKey=#81) Then
		Begin { PgDn }
			Inc(HTMLFirstY, Y2-Y1-30);
			D.NotMore(HTMLFirstY, HTMLMaxY-(Y2-Y1+1));
			HTML_IntDraw(Me, True);
			Recognized:=True
		End;
		If (Key=#0) And (ExtKey=#71) Then
		Begin { Home }
			HTMLFirstY:=0;
			HTML_IntDraw(Me, True);
			Recognized:=True
		End;
		If (Key=#0) And (ExtKey=#79) Then
		Begin { End }
			HTMLFirstY:=HTMLMaxY-(Y2-Y1+1);
			HTML_IntDraw(Me, True);
			Recognized:=True
		End
	End;
	HTML_OnKey:=Recognized
End;




Procedure HTML_OnDestroy(Me:PControl);
Begin
	With Me^, THTML(Data^) Do
	Begin
		If FirstSrcLine<>Nil Then FreeSrcLines;
		If FirstChunk<>Nil Then FreeChunks;
		If FirstPageItem<>Nil Then FreePageItems;

		FreeMem(Data, SizeOf(THTML))
	End
End;

Function HTML_Create(Prog:PProg): PControl;
Var
	Me:PControl;
Begin
	Me:=Kernel.AddCtl(Prog);
	With Me^ Do
	Begin
		OnKey:=HTML_OnKey;
		OnRefocus:=HTML_OnRefocus;
		OnDestroy:=HTML_OnDestroy;

		GetMem(Data, SizeOf(THTML));
		With THTML(Data^) Do
		Begin
			FirstSrcLine:=Nil;
			LastSrcLine:=Nil;
			FirstChunk:=Nil;
			LastChunk:=Nil;
			FirstPageItem:=Nil;
			LastPageItem:=Nil;
			LastPlaceholderItem:=Nil;

			LastWidth:=0;
			HTMLFirstY:=0;

			HTMLScroll:=Scroll_Create(Owner, 0, 0, 0, orVertical,
				HTML_OnScrollUpdate);
			HTMLScroll^.FocusCtl:=Me;
			HTMLScroll^.OwnerCtl:=Me;
			HTMLScroll^.AutoRedraw:=False;
		End
	End;
	HTML_Create:=Me
End;





{ $I}



Procedure Init;
Var
	FName: String;
Begin
	Ini.Select('polit.ini', 'Controls');
	FName:=Ini.GetStr('HTMLFont', 'fonts\dflt.chr');
	MyFont:=Kernel.LoadRes(FName); Kernel.LoadingStep;
End;

Procedure Done;
Begin
	Kernel.FreeRes('dflt.chr');
End;

Begin
End.