{****************************************************************************
 * SoundLib 2 IRQ driver - (c) 2001 by Stefan Goehler                       *
 * Copyright agreement in soundlib.txt - illegal copy without this file!    *
 ****************************************************************************}
{$IFDEF FPC}
{$A+,B-,F+,G+,N-,E-,P+,Q-,R-,T-,V+,X+}
{$ASMMODE INTEL}
{$ELSE}
{$A+,B-,F+,G+,N-,E-,P+,Q-,R-,T-,V+,X+}
{$ENDIF}
unit drv_irq;

interface
{$IFNDEF FPC}
uses dos,ports;
{$ELSE}
uses go32,ports;
{$ENDIF}

type
  tirqhandle = record
    vector     : word;
    irq        : word;
    mask       : word;
    {$IFNDEF FPC}
    oldhandler : pointer;
    handler    : pointer;
    {$ELSE}
    oldhandler : tseginfo;
    handler    : tseginfo;
    {$ENDIF}
  end;

{$IFDEF FPC}
var
  int_ds     : word; external name '___v2prt0_ds_alias';
{$ENDIF}

procedure irq_installhandler(var handle : tirqhandle;irqhnd : pointer;irq : word);
procedure irq_uninstallhandler(var handle : tirqhandle);

implementation

procedure irq_installhandler(var handle : tirqhandle;irqhnd : pointer;irq : word);
begin
  handle.irq := irq;

  with handle do
  begin
    if (irq < 8) then vector := $8 else vector := $70;
    mask    := 1 shl (irq and 7);
    {$IFNDEF FPC}
    handler := irqhnd;
    {$ELSE}
    handler.offset  := irqhnd;
    handler.segment := get_cs;
    {$ENDIF}

    asm cli end;
    {$IFNDEF FPC}
    getintvec(vector+(irq and 7),oldhandler);
    setintvec(vector+(irq and 7),handler);
    {$ELSE}
    get_pm_interrupt(vector+(irq and 7),oldhandler);
    set_pm_interrupt(vector+(irq and 7),handler);
    {$ENDIF}
    if (irq < 8) then
    outp($21,inp($21) and not mask) else
    outp($A1,inp($A1) and not mask);
    asm sti end;
  end;
end;


procedure irq_uninstallhandler(var handle : tirqhandle);
begin
  asm cli end;
  with handle do
  begin
    {$IFNDEF FPC}
    setintvec(vector+(irq and 7),oldhandler);
    {$ELSE}
    set_pm_interrupt(vector+(irq and 7),oldhandler);
    {$ENDIF}
    if (irq < 8) then
    outp($21,inp($21) or mask) else
    outp($A1,inp($A1) or mask);
  end;
  asm sti end;
  fillchar(handle,sizeof(handle),0);
end;

begin
end.
