Unit
	Font; { Polit Halcyon raster font unit }

Interface

Uses
	Log, D, Gfx, Types;

Type
		{ For Raster fonts (RAF) }
	TCharGroup=(cgLatCap, cgLat, cgCyrCap, cgCyr, cgNumbers, cgPunct, cgSpecial);
	TCharGroupSet=Set Of TCharGroup;
	TRFont=Record
		Sig: String[3]; { "RAF" - RAster Font }
		FontWidth, FontHeight: Byte; { CharX - in bytes }
		Groups: TCharGroupSet;
		Width: Array[0..255] Of Byte;
		Data: TMatrix
	End;
		{ For Vector fonts (CHR) }
	TVFontAttr=Set Of (faBold, faItalic, faUnderline);
	TVFont=Record
		Blahblahblah: Array[1..128] Of Byte;
		Sig1: Char; { must be "+" }
		CharsNum: Word;
		Cargo1: Byte;
		FirstChar: Char;
		DefsOffset: Word;
		ScanFlag: Byte; { must be 0 }
		CapTopDist, BaselineDist, DescendantDist: ShortInt;
	End;
		{ Common }
	TFontKind=(fkRaster, fkVector, fkNone);

Var
	DefaultVectorSize: Real;
	DefaultVectorAttr: TVFontAttr;
	BytesPerChar: Integer; { for LPTGFX }

	Function DetectFontKind(P:Pointer): TFontKind;
	Procedure SetFont(P:Pointer);
	Function GetFont: Pointer;
	Procedure Print(X,Y:Integer; Color:LongInt; S:String);
	Procedure PrintVector(X,Y:Integer; Color:TColor; Scale:Real; Attr:TVFontAttr; S:String);
	Procedure PrintFW(X,Y:Integer; Color:LongInt; S:String);
	Function Width(S:String): Integer;
	Function WidthVector(S:String; Scale:Real; Attr:TVFontAttr): Integer;
	Function WidthFW(S:String): Integer;
	Function ChWidth(Ch:Char): Integer;
	Function Height(VFont:Pointer): Integer;
	Function HeightVector(VFnt:Pointer; Scale:Real): Integer;
	Function FileHeightRaster(FName:String): Byte;
	Function BaselineHeightVector(VFnt:Pointer; Scale:Real): Integer;



Implementation

Var
	Fnt: {^TRFont}Pointer;
	FntKind: TFontKind;
	FixedWidth: Integer;

Function DetectFontKind(P:Pointer): TFontKind;
Begin
	If TRFont(P^).Sig='RAF'
		Then DetectFontKind:=fkRaster
		Else If (TVFont(P^).Blahblahblah[1]=Byte('P')) And
			(TVFont(P^).Blahblahblah[2]=Byte('K')) And
			(TVFont(P^).Blahblahblah[3]=8) And
			(TVFont(P^).Blahblahblah[4]=8)
			Then DetectFontKind:=fkVector
			Else DetectFontKind:=fkNone
End;

Procedure SetFont(P:Pointer);
Begin
	Case DetectFontKind(P) Of
		fkRaster: Begin { correct raster font }
			Fnt:=P;
			FntKind:=fkRaster;
			BytesPerChar:=TRFont(Fnt^).FontWidth*TRFont(Fnt^).FontHeight
		End;
		fkVector: Begin { correct vector font }
			Fnt:=P;
			FntKind:=fkVector
		End
		Else Begin
			Fnt:=Nil;
			FntKind:=fkNone;
			Log.ErrMsg('Font', 'SetFont', 'proposed font has no RAF nor PK signature, failed')
		End
	End
End;

Function GetFont: Pointer;
Begin
	GetFont:=Fnt
End;



Procedure PrintRaster(X,Y:Integer; Color:LongInt; S:String);
Var
	ChIndex: Byte;
	CurX, CurXX, CurY: Integer;
	BytesNum: Byte;
	ChX,ChY: Byte;
	CurByte: Byte;
	CurOffs: Word;
	{
		ChIndex - char index in proposed string (cur char - S[ChIndex])
		CurX - current X coord for current char
		CurXX - initial X coord for current byte of current char
			(coords for this byte pixels - CurXX, CurXX+1, ... CurXX+7)
		CurY - current Y coodrd for current byte
		BytesNum - width bytes for cur char
			(FontWidth or less if char is thinner)
		ChX,ChY - byte position for current char image
		CurOffs - current byte offset, CurByte=Data[CurOffs]
		CurByte - current byte of cur char image
	}
Begin
	If FntKind=fkNone Then Exit;

	CurX:=X;
	With TRFont(Fnt^) Do
		For ChIndex:=1 To Length(S) Do
		Begin
			CurOffs:=(Byte(S[ChIndex])-1)*BytesPerChar;
			BytesNum:=D.Min(FontWidth, (Width[Byte(S[ChIndex])]+7) Div 8);
			If BytesNum>0 Then
				For ChX:=0 To BytesNum-1 Do
					For ChY:=0 To FontHeight-1 Do
					Begin
						CurByte:=Data[CurOffs];
						If CurByte<>0 Then
						Begin
							CurXX:=CurX+ChX*8;
							CurY:=Y+ChY;
							If (CurByte And $80)<>0 Then
								Gfx.PutPixel(CurXX+0, CurY, Color);
							If (CurByte And $40)<>0 Then
								Gfx.PutPixel(CurXX+1, CurY, Color);
							If (CurByte And $20)<>0 Then
								Gfx.PutPixel(CurXX+2, CurY, Color);
							If (CurByte And $10)<>0 Then
								Gfx.PutPixel(CurXX+3, CurY, Color);
							If (CurByte And $08)<>0 Then
								Gfx.PutPixel(CurXX+4, CurY, Color);
							If (CurByte And $04)<>0 Then
								Gfx.PutPixel(CurXX+5, CurY, Color);
							If (CurByte And $02)<>0 Then
								Gfx.PutPixel(CurXX+6, CurY, Color);
							If (CurByte And $01)<>0 Then
								Gfx.PutPixel(CurXX+7, CurY, Color)
						End;
						Inc(CurOffs)
					End;
			If FixedWidth=0
				Then Inc(CurX, Width[Byte(S[ChIndex])]+1)
            Else Inc(CurX, FixedWidth)
		End
End;

Procedure PrintVector(X,Y:Integer; Color:TColor; Scale:Real; Attr:TVFontAttr; S:String);
Var
	CurX, I: Integer;
	TempOffs: Word;
	V: Byte;
	C: Char;
	BoldShift: Integer;

	Procedure BoldQualityLine(X1,Y1,X2,Y2,Boldness:Integer; Color:TColor);
	Var
		I: Integer;
	Begin
		For I:=1 To Boldness Do
			Gfx.QualityLine(X1+I-1, Y1, X2+I-1, Y2, Color);
		{Case Boldness Of
			1: Gfx.QualityLine(X1, Y1, X2, Y2, Color);
			2: Begin
				Gfx.QualityLine(X1, Y1, X2, Y2, Color);
				Gfx.QualityLine(X1+1, Y1, X2+1, Y2, Color)
			End;
			Else Begin
				Gfx.QualityLine(X1, Y1, X2, Y2, Color);
				For I:=1 To Boldness Do
					Gfx.Line(X1+I-1, Y1, X2+I-1, Y2, Color);
				Gfx.QualityLine(X1+Boldness-1, Y1, X2+Boldness-1, Y2, Color)
			End
		End}
	End;


	Procedure DrawChar(X,Y:Integer; Color:TColor; Scale:Real; Attr:TVFontAttr; C:Char);
	Var
		TempOffset, CurOffset: Word;
		CurX, CurY: Integer;
		Code1, Code2, Opcode: Byte;
		OpX, OpY: ShortInt;
		Boldness, I: Integer;
	Begin
		If Fnt=Nil Then Exit;
		With TVFont(Fnt^) Do
			If Not D.InRange(Byte(C), Byte(FirstChar),
				Byte(FirstChar)+CharsNum-1) Then
			Begin
				Log.ErrMsg('VFont', 'DrawChar', 'Char not in font: "'+C+'"');
				Exit
			End;

		CurX:=0;
		CurY:=0;
		If faBold In Attr
			Then Boldness:=Round(Scale*2.4)
			Else Boldness:=Round(Scale*1.2);
		TempOffset:=$90+(Integer(C)-Integer(TVFont(Fnt^).FirstChar))*2;
		CurOffset:=(TMatrix(Fnt^)[TempOffset]+TMatrix(Fnt^)[TempOffset+1] Shl 8)+
			$90+TVFont(Fnt^).CharsNum*3;
		Repeat
			Code1:=TMatrix(Fnt^)[CurOffset];
			Code2:=TMatrix(Fnt^)[CurOffset+1];
			Opcode:=(Code1 And $80) Shr 6+(Code2 And $80) Shr 7;
			OpX:=Code1 And $7f;
			If OpX>63 Then OpX:=OpX-128;
			OpY:=Code2 And $7f;
			If OpY>63 Then OpY:=OpY-128;
			If Opcode=3 Then
				BoldQualityLine(X+Round(CurX*Scale), Y-Round(CurY*Scale),
					X+Round(OpX*Scale), Y-Round(OpY*Scale), Boldness, Color);
			CurX:=OpX;
			CurY:=OpY;
			Inc(CurOffset, 2)
		Until Opcode=0
	End;

	Procedure DrawItalicChar(X,Y:Integer; Color:TColor; Scale:Real; Attr:TVFontAttr; C:Char);
	Var
		TempOffset, CurOffset: Word;
		CurX, CurY: Integer;
		Code1, Code2, Opcode: Byte;
		OpX, OpY: ShortInt;
		Boldness, I: Integer;
	Begin
		If Fnt=Nil Then Exit;

		CurX:=0;
		CurY:=0;
		If faBold In Attr
			Then Boldness:=Round(Scale*2.4)
			Else Boldness:=Round(Scale*1.2);
		TempOffset:=$90+(Integer(C)-Integer(TVFont(Fnt^).FirstChar))*2;
		CurOffset:=(TMatrix(Fnt^)[TempOffset]+TMatrix(Fnt^)[TempOffset+1] Shl 8)+
			$90+TVFont(Fnt^).CharsNum*3;
		Repeat
			Code1:=TMatrix(Fnt^)[CurOffset];
			Code2:=TMatrix(Fnt^)[CurOffset+1];
			Opcode:=(Code1 And $80) Shr 6+(Code2 And $80) Shr 7;
			OpX:=Code1 And $7f;
			If OpX>63 Then OpX:=OpX-128;
			OpY:=Code2 And $7f;
			If OpY>63 Then OpY:=OpY-128;
			If Opcode=3 Then
				BoldQualityLine(X+Round(CurX*Scale+CurY*Scale/4), Y-Round(CurY*Scale),
					X+Round(OpX*Scale+OpY*Scale/4), Y-Round(OpY*Scale), Boldness,
					Color);
			CurX:=OpX;
			CurY:=OpY;
			Inc(CurOffset, 2)
		Until Opcode=0
	End;

Begin
	If FntKind=fkNone Then Exit;

	CurX:=X;
	If faUnderline in Attr Then
		For I:=0 To D.Max(0, Round(Scale/1.2)-1) Do
			Gfx.Line(X, Y+HeightVector(Fnt, Scale)-I, X+WidthVector(S, Scale, Attr),
				Y+HeightVector(Fnt, Scale)-I, Color);
	Inc(Y, Round(TVFont(Fnt^).CapTopDist*Scale));
	TempOffs:=$90+TVFont(Fnt^).CharsNum*2;
	If faBold In Attr
		Then BoldShift:=Round(Scale*2.4)-1
		Else BoldShift:=0;
	For V:=1 To Length(S) Do
	Begin
		C:=S[V];
		If faItalic In Attr
			Then DrawItalicChar(CurX, Y, Color, Scale, Attr, C)
			Else DrawChar(CurX, Y, Color, Scale, Attr, C);
		Inc(CurX, Round(TMatrix(Fnt^)[TempOffs+Byte(C)-Integer(TVFont(Fnt^).FirstChar)]*Scale)+BoldShift)
	End
End;

Procedure Print(X,Y:Integer; Color:LongInt; S:String);
Begin
	Case FntKind Of
		fkRaster: PrintRaster(X, Y, Color, S);
		fkVector: PrintVector(X, Y+1, Color, DefaultVectorSize,
			DefaultVectorAttr, S)
	End
End;

Procedure PrintRasterFW(X,Y:Integer; Color:LongInt; S:String);
Begin
	FixedWidth:=TRFont(Fnt^).Width[Byte(' ')]+1;
	Print(X, Y, Color, S);
	FixedWidth:=0
End;

Procedure PrintFW(X,Y:Integer; Color:LongInt; S:String);
Begin
	Case FntKind Of
		fkRaster: PrintRasterFW(X, Y, Color, S);
		fkVector: PrintVector(X, Y+1, Color, DefaultVectorSize,
			DefaultVectorAttr, S)
	End
End;

Function WidthRaster(S:String): Integer;
Var
	Temp, ChIndex: Integer;
Begin
	Temp:=0;
	With TRFont(Fnt^) Do
		For ChIndex:=1 To Length(S) Do
			Inc(Temp, Width[Byte(S[ChIndex])]+1);
	WidthRaster:=Temp
End;

Function WidthVector(S:String; Scale:Real; Attr:TVFontAttr): Integer;
Var
	TempWidth: Integer;
	TempOffs: Word;
	V: Byte;
Begin
	TempOffs:=$90+TVFont(Fnt^).CharsNum*2;
	TempWidth:=0;
	For V:=1 To Length(S) Do
		Inc(TempWidth, Round(TMatrix(Fnt^)[TempOffs+Byte(S[V])-Integer(TVFont(Fnt^).FirstChar)]*Scale));
	If faBold In Attr
		Then Inc(TempWidth, (Round(Scale*2.4)-1)*Length(S));
	WidthVector:=TempWidth
End;

Function Width(S:String): Integer;
Begin
	Case FntKind Of
		fkRaster: Width:=WidthRaster(S);
		fkVector: Width:=WidthVector(S, DefaultVectorSize,
			DefaultVectorAttr);
		fkNone: Width:=0
	End
End;




Function WidthRasterFW(S:String): Integer;
Begin
	WidthRasterFW:=Length(S)*(TRFont(Fnt^).Width[Byte(' ')]+1)
End;

Function WidthFW(S:String): Integer;
Begin
	Case FntKind Of
		fkRaster: WidthFW:=WidthRasterFW(S);
		fkVector: WidthFW:=WidthVector(S, DefaultVectorSize,
			DefaultVectorAttr);
		fkNone: WidthFW:=0
	End
End;






Function ChWidthRaster(Ch:Char): Integer;
Begin
	ChWidthRaster:=TRFont(Fnt^).Width[Byte(Ch)]+1
End;

Function ChWidthVector(Ch:Char; Scale:Real; Attr:TVFontAttr): Integer;
Var
	TempOffs: Word;
	TempWidth: Real;
Begin
	TempOffs:=$90+TVFont(Fnt^).CharsNum*2;
	TempWidth:=TMatrix(Fnt^)[TempOffs+Byte(Ch)-Integer(TVFont(Fnt^).FirstChar)]*Scale;
	If faBold In Attr
		Then TempWidth:=TempWidth+Scale*2.4-1;
	ChWidthVector:=Round(TempWidth)
End;

Function ChWidth(Ch:Char): Integer;
Begin
	Case FntKind Of
		fkRaster: ChWidth:=ChWidthRaster(Ch);
		fkVector: ChWidth:=ChWidthVector(Ch, DefaultVectorSize,
			DefaultVectorAttr);
		fkNone: ChWidth:=0
	End
End;



Function HeightRaster(VFont:Pointer): Integer;
Begin
	HeightRaster:=TRFont(VFont^).FontHeight
End;

Function HeightVector(VFnt:Pointer; Scale:Real): Integer;
Var
	Temp1, Temp2: Integer;
Begin
	Temp1:=TVFont(VFnt^).CapTopDist;
	Temp2:=TVFont(VFnt^).DescendantDist;
	{Height:=Round((TVFont(VFnt^).CapTopDist-TVFont(VFnt^).DescendantDist+1)*Scale);}
	HeightVector:=Round((Temp1-Temp2+1)*Scale)+1
End;

Function Height(VFont:Pointer): Integer;
Begin
	Case DetectFontKind(VFont) Of
		fkRaster: Height:=HeightRaster(VFont);
		fkVector: Height:=HeightVector(VFont, DefaultVectorSize);
		fkNone: Height:=0
	End
End;



Function FileHeightRaster(FName:String): Byte;
Var
	Buf: Array[1..6] Of Byte;
	F: File;
Begin
	Assign(F, FName);
	Reset(F, 1);
	BlockRead(F, Buf[1], 6);
	Close(F);
	FileHeightRaster:=Buf[6]
End;



Function BaselineHeightVector(VFnt:Pointer; Scale:Real): Integer;
Begin
	If VFnt=Nil
		Then BaselineHeightVector:=0
		Else BaselineHeightVector:=Round((TVFont(VFnt^).CapTopDist+1)*Scale)
End;



Begin
	Fnt:=Nil;
	FntKind:=fkNone;
	FixedWidth:=0;
	DefaultVectorSize:=0.65;
	DefaultVectorAttr:=[];

End.