{ ******* MAIN PARSER ******* }

Const
	FSize: Array[1..10] Of Real=(0.65, 0.75, 0.87, 1.0, 1.2, 1.5, 1.8, 2.1, 2.4, 2.7);
	MaxTableCols=30;
	MaxTableRows=500;

Type
	TTableInfo=Record
		Rows, Cols: Integer;
		CurRow, CurCol: Integer;
		Width: Integer;
		OrigX1, OrigY1: Integer;
		CurX1, CurY1: Integer;
		Col: Array[1..MaxTableCols] Of Record
			BGKind: TBGKind;
			BGColor: TColor;
			Width: Integer
		End;
		RowHeight: Array[1..MaxTableRows] Of Integer;
		{Span: Array[1..7, 1..7] Of Byte;
		LastSpanByte: Byte;}
		CellSpacing, CellPadding, Border: Integer;
		TableStart: PPageItem;
		{CurCellBG: TBGKind;
		CurCellBGColor: Integer;}
	End;

Var
	OrigTextState, TextState: TTextAttr;
	CurX, CurY: Integer;
	CurLineHeight, CurBaselineHeight: Integer;

	TagStack: Array[1..30] Of TChunk;
	StackTop: Integer;
	TableStack: Array[1..10] Of ^TTableInfo;
	TableStackTop: Integer;

	CurDocSection: (dsNone, dsTitle, dsBody);
	{LastLineStartIndex: Integer;}
	LastLineStartItem: PPageItem;

{$I Htm_Tool.Inc}

{$I Htm_Tags.Inc}



Procedure ParseChunks(Me:PControl);
Var
	CurChunk: PChunk;
	LastSpace, CurChar: Byte;
	TempStr: String;
	ThisTextHeight, ThisTextBaseline, CurWidth: Integer;
	InsertSoftSpace: Boolean;
	LastFont: Pointer;
Begin
	Log.BugMsg('HTML', 'ParseChunks', 'Parsing started');
	LastFont:=Font.GetFont;
	Font.SetFont(MyFont);

	With DocProps Do
	Begin
		BGColor:=15;
		TextColor:=0;
		LinkColor:=9;
		Title:='';
	End;
	CurDocSection:=dsNone;

	FirstPageItem:=Nil;
	LastPageItem:=Nil;
	LastPlaceholderItem:=Nil;
	StackTop:=0;
	TableStackTop:=0;
	{D.AssignRect(10, 10, Gfx.ScrX-10, Gfx.ScrY-10, Area[1]);}
	{Area[1]:=THTML(Me^.Data^).PageFrame;}
	With Me^.Place Do
	Begin
		Area[1].X1:=3;
		Area[1].Y1:=3;
		Area[1].X2:=X2-X1-2;
		Area[1].Y2:=Y2-Y1-2
	End;
	AreaNum:=1;
	CurX:=Area[1].X1;
	CurY:=Area[1].Y1;
	CurLineHeight:=0;
	CurBaselineHeight:=0;

	HTMLMaxY:=0;

	With OrigTextState Do
	Begin
		Color:=DocProps.TextColor;
		Size:=1;
		Attr:=[];
		Face:=0;
		Align:=alLeft
	End;
	TextState:=OrigTextState;
	LastLineStartItem:=Nil;

		{ ,   :) }
	CurChunk:=FirstChunk;
	While CurChunk<>Nil Do
	Begin
		With CurChunk^ Do
			Case Kind Of
				chTag: Begin { PROCESS TAG }
					Log.BugMsg('HTML', 'ParseChunks', 'Parsing tag...');
					If (Tag='B') Or (Tag='I') Or (Tag='U') Or
						(Tag='BIG') Or (Tag='FONT') Or (Tag='STRONG') Then
						TagCharacterFormatting(CurChunk^);

					If Tag='BR' Then TagBR;
					If Tag='CENTER' Then TagCENTER(CurChunk^);
					If Tag='P' Then TagP(CurChunk^);
					If Tag='HR' Then TagHR(CurChunk^);
					If Tag='IMG' Then TagIMG(CurChunk^);

					If Tag='UL' Then TagUL(CurChunk^);
					If Tag='LI' Then TagLI(CurChunk^);

					If Tag='TABLE' Then TagTABLE(CurChunk^);
					If Tag='TR' Then TagTR(CurChunk^);
					If Tag='TD' Then TagTD(CurChunk^);

					If Tag='TITLE' Then TagTITLE(CurChunk^);
					If Tag='BODY' Then TagBODY(CurChunk^);

				End;

				chText: Case CurDocSection Of
					dsBody: Begin { ADD TEXT ITEM(S) }
						Log.BugMsg('HTML', 'ParseChunks', 'Parsing text...');
						Log.ActMsg('TextState.Size = '+D.St(TextState.Size));
						ThisTextHeight:=Font.HeightVector(MyFont,
							FSize[TextState.Size]);
						Log.ActMsg('VFont.Height OK');
						ThisTextBaseline:=Font.BaselineHeightVector(MyFont,
							FSize[TextState.Size]);
						Log.ActMsg('VFont.BaselineHeight OK');
						TempStr:=String(Text^);
						Log.ActMsg(TempStr);
						Repeat
							If CurX+Font.WidthVector(TempStr, FSize[TextState.Size],
								TextState.Attr)<=Area[AreaNum].X2 Then
							Begin { text fits on current line }
								Log.ActMsg('text fits on current line');
								If ThisTextHeight>CurLineHeight Then
									CurLineHeight:=ThisTextHeight;
								If ThisTextBaseline>CurBaselineHeight Then
									CurBaselineHeight:=ThisTextBaseline;
								AddTextItem(CurX, CurY, CurX+Font.WidthVector(TempStr,
									FSize[TextState.Size], TextState.Attr)-1,
									CurY+ThisTextHeight-1, TempStr, TextState);
								Inc(CurX, Font.WidthVector(TempStr, FSize[TextState.Size],
									TextState.Attr));
									{ check is soft space needed here }
								InsertSoftSpace:=(ssTailing In SoftSpaces) Or
									NextTextChunkHaveLeadingSpace(CurChunk);
								If InsertSoftSpace Then
								Begin
									Inc(CurX, Font.WidthVector(' ',
										FSize[TextState.Size], TextState.Attr));
									If CurX>=Area[AreaNum].X2 Then TagBR
								End;
								TempStr:=''
							End
							Else Begin { text don't fit on current line }
								Log.ActMsg('text don''t fit on current line');
								LastSpace:=0;
								CurChar:=0;
								CurWidth:=0;
								Repeat
									Inc(CurChar);
									Inc(CurWidth, Font.WidthVector(TempStr[CurChar],
										FSize[TextState.Size], TextState.Attr));
									If (TempStr[CurChar]=' ') And
										(CurX+CurWidth<=Area[AreaNum].X2) Then
										LastSpace:=CurChar
								Until (CurWidth>=Area[AreaNum].X2) Or (CurChar>Length(TempStr));
								If LastSpace>0 Then
								Begin { we can break the line on space }
									If ThisTextHeight>CurLineHeight Then
										CurLineHeight:=ThisTextHeight;
									If ThisTextBaseline>CurBaselineHeight Then
										CurBaselineHeight:=ThisTextBaseline;
									AddTextItem(CurX, CurY, CurX+
										Font.WidthVector(Copy(TempStr, 1, LastSpace-1),
										FSize[TextState.Size], TextState.Attr)-1,
										CurY+ThisTextHeight-1, Copy(TempStr, 1, LastSpace-1), TextState);
									TagBR;
									Delete(TempStr, 1, LastSpace)
								End
								Else Begin { line have no spaces before right margin }
									If CurX=Area[AreaNum].X1 Then
									Begin { break inside word }
										CurChar:=0;
										CurWidth:=0;
										Repeat
											Inc(CurChar);
											Inc(CurWidth, Font.WidthVector(TempStr[CurChar],
												FSize[TextState.Size], TextState.Attr))
										Until (CurX+CurWidth>=Area[AreaNum].X2) And (CurChar>1); { <- quick hack }
										Dec(CurWidth, Font.WidthVector(TempStr[CurChar],
											FSize[TextState.Size], TextState.Attr));
										If ThisTextHeight>CurLineHeight Then
											CurLineHeight:=ThisTextHeight;
										If ThisTextBaseline>CurBaselineHeight Then
											CurBaselineHeight:=ThisTextBaseline;
										AddTextItem(CurX, CurY, CurX+CurWidth-1,
											CurY+ThisTextHeight-1, Copy(TempStr, 1, CurChar-1), TextState);
										Delete(TempStr, 1, CurChar-1);
										TagBR
									End
									Else TagBR { transfer to next line }
								End
							End
						Until TempStr=''
					End;
					dsTitle: DocProps.Title:=String(Text^)
				End { text item(s) }
			End; { Case Kind }
		CurChunk:=CurChunk^.Next
	End;
	TagBR;
	Font.SetFont(LastFont)
End;
