Procedure IntAddPageItem(X1,Y1,X2,Y2:Integer);
Begin
	If LastPageItem=Nil Then
	Begin
		New(LastPageItem);
		LastPageItem^.Prev:=Nil;
		LastPageItem^.Next:=Nil;
		FirstPageItem:=LastPageItem
	End
	Else Begin
		New(LastPageItem^.Next);
		LastPageItem^.Next^.Prev:=LastPageItem;
		LastPageItem:=LastPageItem^.Next;
		LastPageItem^.Next:=Nil
	End;
	D.AssignRect(X1, Y1, X2, Y2, LastPageItem^.Place);
	HTMLMaxY:=D.Max(HTMLMaxY, Y2)
End;

Procedure AddTextItem(X1,Y1,X2,Y2:Integer; S: String; VAttr:TTextAttr);
Begin
	Log.BugMsg('HTML', 'AddTextItem', '"'+S+'"');
	If S='' Then
	Begin
		Log.ActMsg('Empty, ignoring');
		Exit
	End;
	IntAddPageItem(X1, Y1, X2, Y2);
	LastPageItem^.Kind:=piText;
	D.AssignPStr(S, LastPageItem^.Text);
	LastPageItem^.Attr:=VAttr
End;

Procedure AddLinkItem(X1,Y1,X2,Y2:Integer; S1, S2: String);
Begin
	IntAddPageItem(X1, Y1, X2, Y2);
	LastPageItem^.Kind:=piLink;
	D.AssignPStr(S1, LastPageItem^.Text);
	D.AssignPStr(S2, LastPageItem^.Href)
End;

Procedure AddImageItem(X1,Y1,X2,Y2:Integer; VImage:Pointer);
Begin
	IntAddPageItem(X1, Y1, X2, Y2);
	LastPageItem^.Kind:=piImage;
	LastPageItem^.Image:=VImage
End;

Procedure AddRectItem(X1,Y1,X2,Y2:Integer; C1,C2:Byte);
Begin
	IntAddPageItem(X1, Y1, X2, Y2);
	LastPageItem^.Kind:=piRect;
	LastPageItem^.Color1:=C1;
	LastPageItem^.Color2:=C2;
End;

Procedure AddBarItem(X1,Y1,X2,Y2:Integer; C:Byte);
Begin
	IntAddPageItem(X1, Y1, X2, Y2);
	LastPageItem^.Kind:=piBar;
	LastPageItem^.Color1:=C;
End;

Procedure RefreshLastPlaceholder;
Var
	Found: Boolean;
Begin
	LastPlaceholderItem:=LastPageItem;
	Found:=False;
	While (LastPlaceholderItem<>Nil) And (Not Found) Do
		If LastPlaceholderItem^.Kind=piPlaceholder
			Then Found:=True
			Else LastPlaceholderItem:=LastPlaceholderItem^.Prev
End;

Procedure PlaceholderToBarItem(X1,Y1,X2,Y2:Integer; C:Byte);
Begin
	If LastPlaceholderItem=Nil Then Exit;

	D.AssignRect(X1, Y1, X2, Y2, LastPlaceholderItem^.Place);
	LastPlaceholderItem^.Kind:=piBar;
	LastPlaceholderItem^.Color1:=C;
	RefreshLastPlaceholder
End;

Procedure AddPlaceholderItem;
Begin
	Log.BugMsg('HTML', 'AddPlaceholderItem', '');
	IntAddPageItem(0, 0, 0, 0);
	LastPageItem^.Kind:=piPlaceholder;
	LastPlaceholderItem:=LastPageItem
End;

Procedure AddTableStartItem;
Begin
	IntAddPageItem(0, 0, 0, 0);
	LastPageItem^.Kind:=piTableStart;
	LastPlaceholderItem:=LastPageItem
End;


{ ********************************************* }

Function FindParam(Tag:TChunk; VParam:String18): String;
Var
	CurParam: PParam;
	Found: Boolean;
Begin
	With Tag Do
		If Kind=chTag Then
		Begin
			VParam:=D.UpCaseLine(VParam);
			CurParam:=FirstParam;
			Found:=False;
			While (CurParam<>Nil) And (Not Found) Do
				If CurParam^.Name=VParam
					Then Found:=True
					Else CurParam:=CurParam^.Next;

			If Found
				Then FindParam:=String(CurParam^.Value^)
				Else FindParam:=''
		End
		Else FindParam:=''
End;

Procedure AddTagToStack(Tag:TChunk);
Begin
	Inc(StackTop);
	TagStack[StackTop]:=Tag
End;

Function FindTopmostTag(VName:String18): Integer;
Var
	I: Integer;
Begin
	FindTopmostTag:=0;
	VName:=D.UpCaseLine(VName);
	For I:=1 To StackTop Do
		If TagStack[I].Tag=VName Then
			FindTopmostTag:=I
End;

Procedure RemoveFromStack(N:Integer);
Var
	I: Integer;
Begin
	For I:=N To StackTop-1 Do
		TagStack[I]:=TagStack[I+1];
	Dec(StackTop)
End;

Procedure RemoveTopmostFromStackByName(VName:String18);
Var
	N: Integer;
Begin
	N:=FindTopmostTag(VName);
	If N<>0 Then
		RemoveFromStack(N)
End;

{ ********************************************* }

Function Color2Int(S:String): Integer;
Const
	XLat: Array['0'..'F'] Of Byte=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
		0, 0, 0, 0, 0, 0, 0,
		10, 11, 12, 13, 14, 15);
Var
	TempR, TempG, TempB: Integer;
Begin
	If S[1]='#' Then Delete (S, 1, 1);
	S:=D.UpCaseLine(S);
	If S='BLACK' Then Color2Int:=0 Else
	If S='RED' Then Color2Int:=12 Else
	If S='GREEN' Then Color2Int:=10 Else
	If S='BLUE' Then Color2Int:=9 Else
	If S='SILVER' Then Color2Int:=7 Else
	Begin
		If Length(S)=6
			Then Color2Int:=Gfx.RGB2Pal[XLat[S[1]], XLat[S[3]], XLat[S[5]]]
			Else Color2Int:=D.Str2Int(S)
	End
End;

Function NextTextChunkHaveLeadingSpace(VChunk:PChunk): Boolean;
Var
	CurChunk: PChunk;
	Found: Boolean;
Begin
	NextTextChunkHaveLeadingSpace:=False;
	Found:=False;
	CurChunk:=VChunk^.Next;
	While (CurChunk<>Nil) And (Not Found) Do
	Begin
		If CurChunk^.Kind=chText Then
		Begin
			NextTextChunkHaveLeadingSpace:=(ssLeading In CurChunk^.SoftSpaces);
			Found:=True
		End;
		CurChunk:=CurChunk^.Next
	End
End;

{ ********************************************* }

Procedure FreePageItems;
Var
	CurItem, NextItem: PPageItem;
Begin
	CurItem:=FirstPageItem;
	While CurItem<>Nil Do
	Begin
		Case CurItem^.Kind Of
			piText: D.FreePStr(CurItem^.Text, 'HTML - FreePageItems');
			piImage: Gfx.FreeImage(CurItem^.Image);
			piLink: Begin
				D.FreePStr(CurItem^.Text, 'HTML - FreePageItems');
				D.FreePStr(CurItem^.Href, 'HTML - FreePageItems')
			End;
		End;
		NextItem:=CurItem^.Next;
		Dispose(CurItem);
		CurItem:=NextItem
	End;
	FirstPageItem:=Nil;
	LastPageItem:=Nil;
End;