{ ************ Label ************ }
Type
	TLabel=Record
		Caption: PStr;
		Align: TAlign;
		Style: TLabelStyle;
		Lines: Integer;
		Picture: Pointer;
	End;

Procedure Label_IntDraw(Me:PControl; VColor:Integer);
Var
	TempStr, CurStr: String;
	TempX, Shift, Pos2, L, CurY, IconWidth: Integer;
Begin
	If ((Not Me^.CanDraw) And (Not Kernel.DontBother)) Or Not Me^.Visible Then
		Exit;
	With Me^, TLabel(Data^), Place Do
	Begin
		If Style=lsBeveled Then
		Begin
			Gfx.Rectangle(X1, Y1, X2, Y2, Kernel.ClrStatic);
			Gfx.Rectangle(X1+1, Y1+1, X2-1, Y2-1, Kernel.ClrWinBG);
			Gfx.Bar(X1+2, Y1+2, X2-2, Y2-2, Kernel.ClrDisabled);
			Shift:=3
		End
		Else
			Shift:=0;
		If Picture<>Nil
			Then IconWidth:=Gfx.ImageX(Picture)+4
			Else IconWidth:=0;
		Font.SetFont(Kernel.CurFont);

		CurStr:=String(Caption^);
		CurY:=Y1+4;
		
		For L:=1 To Lines Do
		Begin
			Pos2:=Pos(#13, CurStr);
			If Pos2=0 Then Pos2:=Length(CurStr)+1;
			TempStr:=Kernel.ShortenStr(Copy(CurStr, 1, Pos2-1),
				X2-X1-IconWidth+1-Shift*2);
			Case Align Of
				alLeft: TempX:=Shift;
				alRight: TempX:=Shift+(X2-X1-IconWidth-Shift*2)-Font.Width(TempStr)+1;
				alCenter: TempX:=Shift+(X2-X1-IconWidth-Shift*2-Font.Width(TempStr)+1) Div 2
			End;
			Font.Print(X1+TempX+IconWidth, CurY, VColor, TempStr);
			If (Picture<>Nil) And (L=1) Then
				Gfx.PutSprite(X1+TempX, CurY-1, Picture);
			Inc(CurY, 14);
			If Pos2<>Length(CurStr)+1 Then
				CurStr:=Copy(CurStr, Pos2+1, Length(CurStr)-Pos2)
		End
	End
End;

Procedure Label_Show(Me:PControl; ToShow:Boolean);
Begin
	With Me^, TLabel(Data^) Do
		Case ToShow Of
			True: Case Style Of
				lsBeveled: Label_IntDraw(Me, Kernel.ClrStatic);
				lsNormal: Label_IntDraw(Me, Kernel.ClrText);
				lsDisabled: Label_IntDraw(Me, Kernel.ClrTextDis);
				lsComment: Label_IntDraw(Me, Kernel.ClrTextCom)
			End;
			False: Begin
				If Style=lsBeveled
					Then Label_IntDraw(Me, Kernel.ClrTextDis)
					Else Label_IntDraw(Me, Kernel.ClrWinBG)
			End
		End
End;


Procedure Label_CountLines(Me:PControl);
Var
	V: Byte;
	TempStr: String;
Begin
	With Me^, TLabel(Data^) Do
	Begin
		Lines:=1;
		TempStr:=String(Caption^);
		For V:=1 To Length(TempStr) Do
			If TempStr[V]=#13 Then
				Inc(Lines)
	End
End;


Procedure Label_Paint(Me:PControl; VX1,VY1,VX2:Integer);
Begin
	{Label_CountLines(Me);}
	With Me^, Owner^.ClientRect, TLabel(Data^) Do
		D.AssignRect(X1+VX1, Y1+VY1, X1+VX2, Y1+VY1+5+Lines*15, Me^.Place);
	Label_Show(Me, True)
End;


Procedure Label_SetCaption(Me:PControl; VCaption:String);
Begin
	With Me^, TLabel(Data^) Do
		If AutoRedraw And Not Kernel.IgnoreAutoRedraw Then
		Begin
			If (Me^.CanDraw Or Kernel.DontBother) And Visible Then
			Begin
				Kernel.BeginPaint(Place);
				Label_Show(Me, False);
				D.ReassignPStr(VCaption, Caption);
				Label_CountLines(Me);
				Label_Show(Me, True);
				Kernel.EndPaint
			End
			Else Begin
				D.ReassignPStr(VCaption, Caption);
				Label_CountLines(Me)
			End
		End
		Else
			D.ReassignPStr(VCaption, Caption)
End;

Procedure Label_SetStyle(Me:PControl; VStyle:TLabelStyle);
Begin
	With Me^, TLabel(Data^) Do
		If AutoRedraw And Not Kernel.IgnoreAutoRedraw Then
		Begin
			If (Me^.CanDraw Or Kernel.DontBother) And Visible Then
			Begin
				Kernel.BeginPaint(Place);
				Label_Show(Me, False);
				Style:=VStyle;
				Label_Show(Me, True);
				Kernel.EndPaint
			End
		End
		Else
			Style:=VStyle
End;

Procedure Label_Draw(Me:PControl);
Begin
	With Me^ Do
		If Me^.CanDraw Or Kernel.DontBother Then
		Begin
			Kernel.BeginPaint(Place);
			If Visible
				Then Label_Show(Me, True)
				Else Kernel.HideCtl(Me);
			Kernel.EndPaint
		End
End;

Function Label_GetCaption(Me:PControl): String;
Begin
	With Me^, TLabel(Data^) Do
		Label_GetCaption:=String(Caption^)
End;

Function Label_GetPicture(Me:PControl): Pointer;
Begin
	With Me^, TLabel(Data^) Do
		Label_GetPicture:=Picture
End;

Procedure Label_CopyToClipboard(Me:PControl);
Var
	F: Text;
Begin
	With Me^, TLabel(Data^) Do
	Begin
		Assign(F, 'clipbrd.txt');
		Rewrite(F);
		Write(F, String(Caption^));
		Close(F);
		If D.HostOS=osWindows Then
			D.WinCopy
	End
End;

Procedure Label_OnDestroy(Me:PControl);
Begin
	With Me^ Do
	Begin
		D.FreePStr(TLabel(Data^).Caption, 'Ctl - Label_Destroy');
		FreeMem(Data, SizeOf(TLabel))
	End
End;

Function Label_Create(Prog:PProg; VCaption:String; VPicture:Pointer;
	VStyle:TLabelStyle; VAlign:TAlign): PControl;
Var
	Me: PControl;
Begin
	Me:=Kernel.AddCtl(Prog);
	With Me^ Do
	Begin
		OnDestroy:=Label_OnDestroy;
		Enabled:=False;

		GetMem(Me^.Data, SizeOf(TLabel));
		With TLabel(Data^) Do
		Begin
			D.AssignPStr(VCaption, Caption);
			Label_CountLines(Me);
			Picture:=VPicture;
			Style:=VStyle;
			Align:=VAlign
		End
	End;
	Label_Create:=Me
End;
