{ ************ Memo, second chance ************ }
Type
	TTextLineKind=(tlNormal, {tlBroken,} tlParaEnd);
	PTextLine=^TTextLine;
	TTextLine=Record
		Data: PStr;
		Kind: TTextLineKind;
		Prev, Next: PTextLine
	End;

	TMemo=Record
		FirstLine: PTextLine;
		Kind: TMemoKind;
		LastWidth: Integer;
		MemoScroll: PControl;
		FirstVisLine, TotalLines: Integer;
		CursorLine, CursorChar, CursorX: Integer;
		CursorColor: TColor;
		CursorLinePtr: PTextLine;
		SelStartLine, SelStartChar, SelEndLine, SelEndChar: Integer;
		SelPresent:Boolean;
		VisLen:Integer;
		MyFont: Pointer;
		MyFontHeight: Integer;
		ContextMenu: PMenuItem;
		Coding: Integer;

		ColorBG, ColorFont: TColor;
	End;

Var
	MemoFastScroll: Boolean;
	Recode, RecodeBack: Array[1..2, 0..255] Of Char;
	AlwaysTextCursor: Boolean;




Function Memo_LineByIndex(Me:PControl; Index:Integer): PTextLine;
Var
	CurLine: PTextLine;
	LineNum: Integer;
Begin
	With Me^, TMemo(Data^), Place Do
	Begin
		LineNum:=1;
		CurLine:=FirstLine;
		While (CurLine<>Nil) And (LineNum<>Index) Do
		Begin
			Inc(LineNum);
			CurLine:=CurLine^.Next
		End;
		If CurLine=Nil
			Then Memo_LineByIndex:=Nil
			Else Memo_LineByIndex:=CurLine
	End
End;

Function Memo_IndexByLine(Me:PControl; VLine:PTextLine): Integer;
Var
	CurLine: PTextLine;
	LineNum: Integer;
Begin
	With Me^, TMemo(Data^), Place Do
	Begin
		LineNum:=1;
		CurLine:=FirstLine;
		While (CurLine<>Nil) And (CurLine<>VLine) Do
		Begin
			Inc(LineNum);
			CurLine:=CurLine^.Next
		End;
		If CurLine=Nil
			Then Memo_IndexByLine:=-1
			Else Memo_IndexByLine:=LineNum
	End
End;

Procedure Memo_CountLines(Me:PControl);
Var
	CurLine: PTextLine;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		TotalLines:=0;
		CurLine:=FirstLine;
		While CurLine<>Nil Do
		Begin
			Inc(TotalLines);
			CurLine:=CurLine^.Next
		End
	End
End;


Procedure Memo_ShowLines(Me:PControl; VFirst,VNum:Integer; ToClear:Boolean);
Var
	CurLine: PTextLine;
	CurY, CurLineNum: Integer;
	SelX1, SelX2: Integer;
	ClrFont: TColor;
Begin
	With Me^, TMemo(Data^), Place Do
	Begin
		CurY:=Y1+4+(VFirst-FirstVisLine)*MyFontHeight;
		CurLine:=Memo_LineByIndex(Me, VFirst);
		Font.SetFont(MyFont);
		If Kind=mkEditable
			Then ClrFont:=Kernel.ClrEditFont
			Else ClrFont:=Kernel.ClrStatic;
		{TotalLines:=Memo_CountLines(Me);}
		For CurLineNum:=VFirst To D.Min3(VFirst+VNum-1,
			FirstVisLine+VisLen-1, TotalLines) Do
		Begin
			If CurLineNum>=FirstVisLine Then
			Begin
				If ToClear Then
					Gfx.Bar(X1+3, CurY, X2-3, CurY+MyFontHeight-1, ColorBG);
				If SelPresent And (SelStartLine<=CurLineNum) And
					(SelEndLine>=CurLineNum) Then
				Begin
					If SelStartLine=CurLineNum
						Then SelX1:=X1+3+Font.Width(Copy(String(CurLine^.Data^),
							1, SelStartChar))
						Else SelX1:=X1+3;
					If SelEndLine=CurLineNum
						Then SelX2:=X1+3+Font.Width(Copy(String(CurLine^.Data^),
							1, SelEndChar))
						Else SelX2:=X1+3+Font.Width(String(CurLine^.Data^));
					Gfx.Bar(SelX1, CurY, SelX2, CurY+MyFontHeight-1,
						Kernel.ClrEditSel)
				End;
				Font.Print(X1+4, CurY, ClrFont, String(CurLine^.Data^))
			End;
			Inc(CurY, MyFontHeight);
			CurLine:=CurLine^.Next
		End
	End
End;

Procedure Memo_DrawCursor(Me:PControl);
Var
	CurY: Integer;
Begin
	With Me^, TMemo(Data^), Place Do
	Begin
		CurY:=Y1+4+(CursorLine-FirstVisLine)*MyFontHeight;
		Kernel.BeginPaintCoord(CursorX, CurY, CursorX, CurY+MyFontHeight-1);
		Gfx.Line(CursorX, CurY, CursorX, CurY+MyFontHeight-1, CursorColor);
		Kernel.EndPaint
	End
End;

Procedure Memo_ShowCursor(Me:PControl; Show:Boolean);
Var
	RealColorBG: TColor;
	Sel1, Sel2: Integer;
Begin
	With Me^, Place, TMemo(Data^) Do
	Begin
		If Not D.InRange(CursorLine, FirstVisLine,
			FirstVisLine+VisLen-1) Then Exit;

		If SelPresent Then
		Begin
			If (SelStartLine<=CursorLine) And (SelEndLine>=CursorLine) Then
			Begin
				If SelStartLine=CursorLine
					Then Sel1:=SelStartChar
					Else Sel1:=0;
				If SelEndLine=CursorLine
					Then Sel2:=SelEndChar
					Else Sel2:=Length(String(CursorLinePtr^.Data^));
				If D.InRange(CursorChar, Sel1, Sel2)
					Then RealColorBG:=Kernel.ClrEditSel
					Else RealColorBG:=ColorBG
			End
			Else
				RealColorBG:=ColorBG
		End
		Else
			RealColorBG:=ColorBG;
		If Show
			Then CursorColor:=ColorFont
			Else CursorColor:=RealColorBG;

		Memo_DrawCursor(Me)
	End
End;

Procedure Memo_RefreshCursorX(Me:PControl);
Begin
	With Me^, TMemo(Data^), Place Do
	Begin
		Font.SetFont(MyFont);
		CursorX:=X1+3+Font.Width(Copy(String(CursorLinePtr^.Data^), 1, CursorChar));
		D.ToRange(CursorX, X1+3, X2-3)
	End
End;


Procedure Memo_Reformat(Me:PControl); Forward;

Procedure Memo_IntDraw(Me:PControl; DrawScroll:Boolean);
Var
	CurLine: PTextLine;
	CurY: Integer;
	CurLineNum: Integer;
	Temp: String;
	V, I: Byte;
Begin
	If ((Not Me^.CanDraw) And (Not Kernel.DontBother)) Or Not Me^.Visible Then
		Exit;

	With Me^, TMemo(Data^), Place Do
	Begin
		If X2-X1-8<>LastWidth Then
		Begin
			Memo_Reformat(Me);
			LastWidth:=X2-X1-8
		End;
		Memo_CountLines(Me);
		VisLen:=(Y2-Y1-6) Div MyFontHeight;
		D.ToRange(FirstVisLine, 1, TotalLines-VisLen+1);
		If CursorLine>TotalLines Then
		Begin
			CursorLine:=TotalLines;
			CursorChar:=0;
			CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);
			Memo_RefreshCursorX(Me)
		End;

		If Kind=mkEditable Then
		Begin
			Gfx.AngleRect(X1, Y1, X2, Y2, Kernel.ClrEditDarker);
			Gfx.AngleRect(X1+1, Y1+1, X2, Y2, Kernel.ClrEditDark);
			Gfx.Bar(X1+2, Y1+2, X2, Y2, Kernel.ClrEdit)
		End
		Else Begin { Kind=mkReadOnly }
			Gfx.RectAngle(X1, Y1, X2, Y2, Kernel.ClrStatic);
			Gfx.RectAngle(X1+1, Y1+1, X2-1, Y2-1, Kernel.ClrWinBG);
			Gfx.Bar(X1+2, Y1+2, X2-2, Y2-2, Kernel.ClrDisabled)
		End;

		Memo_ShowLines(Me, FirstVisLine, VisLen, False);
		Memo_RefreshCursorX(Me);

		If DrawScroll Then
		Begin
			TControl(MemoScroll^).Enabled:=Enabled;
			Scroll_SetParams(MemoScroll, TotalLines, VisLen, FirstVisLine);
			If (MemoScroll^.CanDraw Or Kernel.DontBother) And Visible Then
				Scroll_PaintDirect(MemoScroll, X2+2, Y1, X2+16, Y2)
		End
	End
End;

Procedure Memo_Scroll(Me:PControl; Delta:Integer; DrawScroll:Boolean);
Var
	I:Integer;
	P:Pointer;
	Size:Word;
Begin
	With Me^, Place, TMemo(Data^) Do
	Begin
		If (Abs(Delta)>VisLen) Or MemoFastScroll Then
		Begin
			Inc(FirstVisLine, Delta);
			Kernel.BeginPaint(Place);
			Memo_IntDraw(Me, DrawScroll);
			Kernel.EndPaint;
			Exit
		End;

		Size:=Gfx.ImageSize(X2-X1-7, MyFontHeight);
		GetMem(P, Size);
		Memo_ShowCursor(Me, False);
		Kernel.BeginPaint(Place);

		If Delta>0 Then
		Begin { moving down, text shifts up }
			For I:=Delta To VisLen-1 Do
			Begin
				Gfx.GetImage(X1+4, Y1+4+I*MyFontHeight, X2-4,
					Y1+4+(I+1)*MyFontHeight-1, P);
				Gfx.PutImage(X1+4, Y1+4+(I-Delta)*MyFontHeight, P)
			End;
			Inc(FirstVisLine, Delta);
			Memo_ShowLines(Me, FirstVisLine+VisLen-Delta, Delta, True)
		End
		Else Begin { moving up, text shifts down }
			Delta:=-Delta;
			For I:=VisLen-1 DownTo Delta Do
			Begin
				Gfx.GetImage(X1+4, Y1+4+(I-Delta)*MyFontHeight, X2-4,
					Y1+4+(I-Delta+1)*MyFontHeight-1, P);
				Gfx.PutImage(X1+4, Y1+4+I*MyFontHeight, P)
			End;
			Dec(FirstVisLine, Delta);
			Memo_ShowLines(Me, FirstVisLine, Delta, True)
		End;
		FreeMem(P, Size);

		If DrawScroll Then
		Begin
			TControl(MemoScroll^).Enabled:=Enabled;
			Scroll_SetParams(MemoScroll, TotalLines, VisLen, FirstVisLine);
			If (MemoScroll^.CanDraw Or Kernel.DontBother) And Visible Then
				Scroll_PaintDirect(MemoScroll, X2+2, Y1, X2+16, Y2)
		End;
		Kernel.EndPaint;
		Memo_ShowCursor(Me, True)
	End
End;


Procedure Memo_Paint(Me:PControl; VX1, VY1, VX2, VY2:Integer);
Begin
	With Me^.Owner^.ClientRect Do
		D.AssignRect(X1+VX1, Y1+VY1, X1+VX2-16, Y1+VY2, Me^.Place);
	Memo_IntDraw(Me, True)
End;

Procedure Memo_TryAutoRedraw(Me:PControl);
Begin
	If Me^.AutoRedraw And Not Kernel.IgnoreAutoRedraw Then
		If (Me^.CanDraw Or Kernel.DontBother) And Me^.Visible Then
		Begin
			Kernel.BeginPaint(Me^.Place);
			Memo_IntDraw(Me, True);
			Kernel.EndPaint
		End
End;

Procedure Memo_Draw(Me:PControl);
Begin
	With Me^ Do
		If CanDraw Or Kernel.DontBother Then
		Begin
			Kernel.BeginPaint(Me^.Place);
			If Visible
				Then Memo_IntDraw(Me, True)
				Else Kernel.HideCtl(Me);
			Kernel.EndPaint
		End
End;


Procedure Memo_Clear(Me:PControl);
Var
	CurLine, NextLine: PTextLine;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		CurLine:=FirstLine;
		While CurLine<>Nil Do
		Begin
			D.FreePStr(CurLine^.Data, 'Ctl - Memo_Clear');
			NextLine:=CurLine^.Next;
			Dispose(CurLine);
			CurLine:=NextLine
		End;

		New(FirstLine);
		D.AssignPStr('', FirstLine^.Data);
		FirstLine^.Kind:=tlParaEnd;
		FirstLine^.Prev:=Nil;
		FirstLine^.Next:=Nil;
		TotalLines:=1;
		FirstVisLine:=1;
		CursorLine:=1;
		CursorChar:=0;
		CursorLinePtr:=FirstLine;
		Coding:=0;

		Memo_TryAutoRedraw(Me)
	End;
End;




Procedure Memo_FormatParaEnd(Me:PControl; OrigLine:PTextLine; Buffer:Pointer; BufSize:Word);
Const
	TabStr='   ';
Var
	BufCur: Word;
	CurLine, GlobalNextLine: PTextLine;
	V: Byte;
	EndOfPara: Boolean;
	TempStr: String;
	Ch: Char;
	TempWidth: Integer;
	ControlWidth: Integer;

	Function ReadChar:Char;
	Begin
		ReadChar:=Char(TMatrix(Buffer^)[BufCur]);
		Inc(BufCur)
	End;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		If BufSize=0 Then {     :( }
		Begin
			D.AssignPStr('', OrigLine^.Data);
			OrigLine^.Kind:=tlParaEnd;
			If OrigLine^.Next<>Nil Then
				OrigLine^.Next^.Prev:=OrigLine;
			Exit
		End;
			{  ஧  i   浪 }
		EndOfPara:=False;

			{  i    OrigLine^.Data=Nil }
		GlobalNextLine:=OrigLine^.Next;
		CurLine:=OrigLine;

		BufCur:=0;
		If Kernel.IgnoreAutoRedraw { starting }
			Then ControlWidth:=300
			Else ControlWidth:=Place.X2-Place.X1-8;
		Repeat
			Font.SetFont(MyFont);
			TempStr:='';
			TempWidth:=0;
			
			While (BufCur<BufSize) And (Length(TempStr)<255) And Not
				(TempWidth>ControlWidth) Do
			Begin
				Ch:=ReadChar;
				Case Ch Of
					#9:Begin
						TempStr:=TempStr+TabStr;
						Inc(TempWidth, Font.Width(TabStr))
					End;
					#13, #10, #0: { do nothing } ;
					Else Begin
						TempStr:=TempStr+Ch;
						Inc(TempWidth, Font.ChWidth(Ch))
					End
				End
			End;

			If TempWidth>ControlWidth Then
			Begin
				V:=Length(TempStr);
				Repeat
					Dec(TempWidth, Font.ChWidth(TempStr[V]));
					Dec(V)
				Until ((TempStr[V]=' ') And (TempWidth<=ControlWidth)) Or
					(V=0);
				If V=0 Then
				Begin
					V:=Length(TempStr);
					TempWidth:=Font.Width(TempStr);
					Repeat
						Dec(TempWidth, Font.ChWidth(TempStr[V]));
						Dec(V)
					Until TempWidth<=ControlWidth
				End;
				D.AssignPStr(Copy(TempStr, 1, V), CurLine^.Data);
				CurLine^.Kind:=tlNormal;
				Dec(BufCur, Length(TempStr)-V);
				New(CurLine^.Next);
				CurLine^.Next^.Prev:=CurLine;
				CurLine^.Next^.Next:={Nil}GlobalNextLine;
				CurLine:=CurLine^.Next;
			End
			Else Begin
				D.AssignPStr(TempStr, CurLine^.Data);
				EndOfPara:=True;
				CurLine^.Kind:=tlParaEnd;
				CurLine^.Next:=GlobalNextLine;
				If GlobalNextLine<>Nil  Then
					GlobalNextLine^.Prev:=CurLine
			End
		Until EndOfPara
	End
End;

Procedure Memo_UnformatParaEnd(Me:PControl; OrigLine:PTextLine; Var Buffer:Pointer; Var BufSize:Word);
Var
	CurLine, NextLine: PTextLine;
	TempSize, TempLen: Word;
	V: Byte;
	EndOfPara: Boolean;
	Finished, MoreThan1Line: Boolean;
Begin
	With Me^, TMemo(Data^) Do
	Begin
			{ ᫨ ஧i ⥪  ࠣi }
		CurLine:=OrigLine;
		TempSize:=0;
		Finished:=False;
		Repeat
			Inc(TempSize, Length(String(CurLine^.Data^)));
			If CurLine^.Kind=tlParaEnd
				Then Finished:=True
				Else CurLine:=CurLine^.Next
		Until Finished;
		GetMem(Buffer, TempSize);
			{ ९  筨    i i쭨 }
		BufSize:=0;
		CurLine:=OrigLine;
		Finished:=False;
		MoreThan1Line:=False;
		While Not Finished Do
		Begin
			TempLen:=Length(String(CurLine^.Data^));
			Move(String(CurLine^.Data^)[1], TMatrix(Buffer^)[BufSize], TempLen);
			Inc(BufSize, TempLen);
			D.FreePStr(CurLine^.Data, 'Ctl - Memo_UnformatParaEnd');
			If (CurLine^.Kind=tlParaEnd) Then
				Finished:=True;
			NextLine:=CurLine^.Next;
			If CurLine=OrigLine
				Then CurLine^.Data:=Nil
				Else Dispose(CurLine);
			If Not Finished Then
			Begin
				MoreThan1Line:=True;
				CurLine:=NextLine
			End
		End;
		If MoreThan1Line Then 
		Begin
			If CurLine<>Nil Then { runtime 216 after resizing MemoTest ! }
			Begin
				OrigLine^.Next:={CurLine^.Next}NextLine;
				If {CurLine^.Next}NextLine<>Nil Then
					{CurLine^.Next^}NextLine^.Prev:=OrigLine
			End
			Else
				OrigLine^.Next:=Nil
		End
	End
End;


{ 24.11.02 } { 26.11.02 }
Procedure Memo_PosXYToOffset(Me:PControl; VLine,VChar:Integer;
	Var Offset,ParaEnds:LongInt);
Var
	CurLine: PTextLine;
	CurLineNum: Integer;
	CurOffset, CurParaEnds: LongInt;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		CurLine:=FirstLine;
		CurLineNum:=1;
		CurOffset:=0;
		CurParaEnds:=0;
		Font.SetFont(MyFont);
		While CurLineNum<>VLine Do
		Begin
			Inc(CurOffset, Length(String(CurLine^.Data^)));
			If CurLine^.Kind=tlParaEnd Then
				Inc(CurParaEnds);
			Inc(CurLineNum);
			CurLine:=CurLine^.Next
		End;
		Inc(CurOffset, VChar)
	End;
	Offset:=CurOffset;
	ParaEnds:=CurParaEnds
End;

Procedure Memo_PosOffsetToXY(Me:PControl; VOffset,VParaEnds:LongInt;
	Var RLine,RChar:Integer);
Var
	CurLine: PTextLine;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		RLine:=1;
		CurLine:=FirstLine;
		While Not ((VOffset<=Length(String(CurLine^.Data^))) And
			(VParaEnds=0)) Do
		Begin
			Dec(VOffset, Length(String(CurLine^.Data^)));
			If CurLine^.Kind=tlParaEnd Then
				Dec(VParaEnds);
			Inc(RLine);
			CurLine:=CurLine^.Next
		End;
		RChar:=VOffset
	End
End;


Procedure Memo_IntReformatParaEnd(Me:PControl; VLine:PTextLine);
Var
	Buffer: Pointer;
	BufSize: Word;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		If Not ((String(VLine^.Data^)='') And (VLine^.Kind=tlParaEnd)) Then
		Begin
			Memo_UnformatParaEnd(Me, VLine, Buffer, BufSize);
			Memo_FormatParaEnd(Me, VLine, Buffer, BufSize);
			FreeMem(Buffer, BufSize)
		End
	End
End;

Procedure Memo_ReformatParaEnd(Me:PControl; VLine:PTextLine);
Var
	CursorOffset, SelStartOffset, SelEndOffset: LongInt;
	CursorPE, SelStartPE, SelEndPE: LongInt;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		Log.BugMsg('Memo', 'ReformatParaEnd', 'Starting...');
		Log.ActMsg('First line: '+String(FirstLine^.Data^));
		If Kind=mkEditable Then
		Begin
			Memo_PosXYToOffset(Me, CursorLine, CursorChar, CursorOffset, CursorPE);
			If SelPresent Then
			Begin
				Memo_PosXYToOffset(Me, SelStartLine, SelStartChar, SelStartOffset, SelStartPE);
				Memo_PosXYToOffset(Me, SelEndLine, SelEndChar, SelEndOffset, SelEndPE)
			End;
			Log.BugMsg('Memo', 'ReformatParaEnd', 'Kind=mkEditable, saved cursor position');
		End;

		Memo_IntReformatParaEnd(Me, VLine);
		Log.BugMsg('Memo', 'ReformatParaEnd', 'IntReformat called');
		Memo_CountLines(Me);
		Log.BugMsg('Memo', 'ReformatParaEnd', 'CountLines called');

		If Kind=mkEditable Then
		Begin
			Memo_PosOffsetToXY(Me, CursorOffset, CursorPE, CursorLine, CursorChar);
			CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);
			Memo_RefreshCursorX(Me);
			If SelPresent Then
			Begin
				Memo_PosOffsetToXY(Me, SelStartOffset, SelStartPE, SelStartLine, SelStartChar);
				Memo_PosOffsetToXY(Me, SelEndOffset, SelEndPE, SelEndLine, SelEndChar)
			End;
			Log.BugMsg('Memo', 'ReformatParaEnd', 'Kind=mkEditable, restored cursor position');
		End
	End
End;


Procedure Memo_Reformat(Me:PControl);
Var
	CurLine: PTextLine;
	StartOffset, CursorOffset, SelStartOffset, SelEndOffset: LongInt;
	StartPE, CursorPE, SelStartPE, SelEndPE: LongInt;
	TempInt: Integer;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		Memo_PosXYToOffset(Me, FirstVisLine, 0, StartOffset, StartPE);

		If Kind=mkEditable Then
		Begin
			Memo_PosXYToOffset(Me, CursorLine, CursorChar, CursorOffset, CursorPE);
			If SelPresent Then
			Begin
				Memo_PosXYToOffset(Me, SelStartLine, SelStartChar, SelStartOffset, SelStartPE);
				Memo_PosXYToOffset(Me, SelEndLine, SelEndChar, SelEndOffset, SelEndPE)
			End
		End;

		CurLine:=FirstLine;
		Memo_IntReformatParaEnd(Me, CurLine);
		Repeat
			{Log.BugMsg('Ctl', 'Memo_Reformat', String(CurLine^.Data^)+
				' '+D.StL(MemAvail));}
			If (CurLine^.Kind=tlParaEnd) And (CurLine^.Next<>Nil) Then
				Memo_IntReformatParaEnd(Me, CurLine^.Next);
			CurLine:=CurLine^.Next;
		Until CurLine=Nil;
		Memo_CountLines(Me);

		If Kind=mkEditable Then
		Begin
			Memo_PosOffsetToXY(Me, CursorOffset, CursorPE, CursorLine, CursorChar);
			CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);
			Memo_RefreshCursorX(Me);
			If SelPresent Then
			Begin
				Memo_PosOffsetToXY(Me, SelStartOffset, SelStartPE, SelStartLine, SelStartChar);
				Memo_PosOffsetToXY(Me, SelEndOffset, SelEndPE, SelEndLine, SelEndChar)
			End
		End;

		Memo_PosOffsetToXY(Me, StartOffset, StartPE, FirstVisLine, TempInt);
		D.ToRange(FirstVisLine, 1, TotalLines-VisLen+1)
	End;
End;








Procedure Memo_ApplyCoding(Me:PControl; Buffer:Pointer; BufSize:Word;
	Coding:Integer);
Var
	I: Integer;
Begin
	With TMemo(Me^.Data^) Do
		For I:=0 To BufSize-1 Do
			TMatrix(Buffer^)[I]:=Byte(Recode[Coding, TMatrix(Buffer^)[I]]);
End;

Function Memo_DetectCoding(Buffer:Pointer; BufSize:Word): Integer;
Const
	ScoredChar: String='';
Type
	TFreq=Record
		C: Char;
		N: LongInt
	End;
Var
	Freq: Array[0..2, 0..255] Of TFreq;
	TempFreq: TFreq;
	I, TotalChars: Integer;
	V, V2: Byte;
	Found: Boolean;
	CurLine: PTextLine;
	TempStr: String;
	Score: Array[0..2] Of Integer;
Begin
		{ 㢠  }
	For V:=0 To 255 Do
	Begin
		Freq[0, V].C:=Char(V);
		Freq[0, V].N:=0;
		Freq[1, V].C:=Char(V);
		Freq[1, V].N:=0;
		Freq[2, V].C:=Char(V);
		Freq[2, V].N:=0
	End;
	If BufSize>7000 Then
		BufSize:=7000;
	For I:=0 To BufSize-1 Do
	Begin
		Inc(Freq[0, TMatrix(Buffer^)[I]].N);
		Inc(Freq[1, Byte(Recode[1, TMatrix(Buffer^)[I]])].N);
		Inc(Freq[2, Byte(Recode[2, TMatrix(Buffer^)[I]])].N)
	End;

		{ 㢠  }
	For I:=0 To 2 Do
		For V:=0 To 254 Do
			For V2:=V+1 To 255 Do
				If Freq[I, V].N<Freq[I, V2].N Then
				Begin
					TempFreq:=Freq[I, V];
					Freq[I, V]:=Freq[I, V2];
					Freq[I, V2]:=TempFreq
				End;

		{ 筥 i襭 }
	For I:=0 To 2 Do
	Begin
		Score[I]:=0;
		For V:=1 To Length(ScoredChar) Do
		Begin
			Found:=False;
			V2:=0;
			Repeat
				If Freq[I, V2].C=ScoredChar[V]
					Then Found:=True
					Else Inc(V2)
			Until Found Or (V2=255);
			Inc(Score[I], 255-V2)
		End
	End;

	{If Score[1]>Score[0]
		Then Memo_DetectCoding:=True
		Else Memo_DetectCoding:=False}
	If (Score[0]>Score[1]) And (Score[0]>Score[2])
		Then Memo_DetectCoding:=0
		Else If Score[1]>Score[2]
			Then Memo_DetectCoding:=1
			Else Memo_DetectCoding:=2
End;








Procedure Memo_LoadFile(Me:PControl; FName:String; ToClear:Boolean);
Const
	MaxBufSize=16384{2048};
Var
	Buffer: Pointer;
	BufStart, BufEnd, BufSize, TempSize: Word;
	F:File;
	CurLine:PTextLine;
	TempPtr:Pointer;
	FirstBlockLine, NextLine: PTextLine;
	TempStr:String;
	LastLineParaEnd:Boolean;
	TempResult:(trOK, trBufferEnd, trLineFeed);
Begin
	(* 08.11.2002 code segment too large, temporary commenting *)
	With Me^, TMemo(Data^) Do
	Begin
		If Not D.FileExist(FName) Then {  i 䠩? }
		Begin
			Log.ErrMsg('Ctl', 'Memo_LoadFile', 'File not found: '+FName);
			Log.ActMsg('Ignoring load request');
			Exit
		End;
		Assign(F, FName);
		FileMode:=0;
		Reset(F, 1);
		FileMode:=2;
		If FileSize(F)=0 Then {   ஦i i? }
		Begin
			Log.ErrMsg('Ctl', 'Memo_LoadFile', 'File is empty (length=0): '+FName);
			Log.ActMsg('Ignoring load request');
			Close(F);
			If ToClear Then
			Begin
				Memo_Clear(Me);
				Memo_TryAutoRedraw(Me)
			End;
			Exit
		End;
		If (FileSize(F)*1.7)>MemAvail Then
		Begin
			Log.ErrMsg('Ctl', 'Memo_LoadFile', 'Probably not enough memory: '+FName);
			Log.ActMsg('Ignoring load request');
			Close(F);
			If ToClear Then
			Begin
				Memo_Clear(Me);
				Memo_TryAutoRedraw(Me)
			End;
			Exit
		End;

			{ ਣ㢠  }
		New(CurLine);
		FirstBlockLine:=CurLine;
		CurLine^.Data:=Nil;
		CurLine^.Next:=Nil;
		GetMem(Buffer, MaxBufSize);
		BlockRead(F, Buffer^, MaxBufSize, BufSize);
		BufStart:=0;
		If ToClear Then
		Begin
			Coding:=Memo_DetectCoding(Buffer, BufSize);
			If Coding<>0 Then
				Memo_ApplyCoding(Me, Buffer, BufSize, Coding)
		End
		Else
			Coding:=0;
			{ 横  栬 }
		While BufStart<BufSize Do
		Begin

				{ ⢮   }
			If BufStart>0 Then
			Begin
				New(CurLine^.Next);
				CurLine^.Next^.Prev:=CurLine;
				CurLine^.Next^.Next:=Nil;
				CurLine^.Next^.Data:=Nil; {  Memo_FormatParaEnd }
				CurLine:=CurLine^.Next
			End;
				{ 㪠  i 㯭 0d0ah }
			BufEnd:=BufStart;
			TempResult:=trOK;
			Repeat
				If BufEnd=BufSize
					Then TempResult:=trBufferEnd
					Else If TMatrix(Buffer^)[BufEnd] In [$0d, $0a]
						Then TempResult:=trLineFeed
						Else Inc(BufEnd)
			Until TempResult<>trOK;

			If TempResult=trBufferEnd Then
			Begin { i諨  i  }
				LastLineParaEnd:=False;
				If Not EOF(F) Then
				Begin { ॡ ⠦ }
					Move(TMatrix(Buffer^)[BufStart], Buffer^, BufEnd-BufStart);
					BlockRead(F, TMatrix(Buffer^)[BufEnd-BufStart],
						MaxBufSize-(BufEnd-BufStart), TempSize);
					If Coding<>0 Then
						Memo_ApplyCoding(Me,
							Addr(TMatrix(Buffer^)[BufEnd-BufStart]),
							TempSize, Coding);
					BufSize:=(BufEnd-BufStart)+TempSize;
					BufStart:=0
				End
				Else Begin { ⠭i 冷 䠩 }
					TempPtr:=Addr(TMatrix(Buffer^)[BufStart]);
					Memo_FormatParaEnd(Me, CurLine, TempPtr, BufEnd-BufStart);
					BufStart:=BufEnd+2;
					While CurLine^.Kind<>tlParaEnd Do
						CurLine:=CurLine^.Next;
				End
			End
			Else Begin {  ଠ쭮,  i  ॢ 浪 }
				LastLineParaEnd:=True;
				Memo_FormatParaEnd(Me, CurLine,
					Addr(TMatrix(Buffer^)[BufStart]), BufEnd-BufStart);

				While CurLine^.Kind<>tlParaEnd Do
					CurLine:=CurLine^.Next;
				BufStart:=BufEnd+2
			End
		End;
		Close(F);
		FreeMem(Buffer, MaxBufSize);

			{ ஧i ⠦  ॡ }
		If ToClear Then
		Begin {  ,  㫮  Memo  쮣 }
			Memo_Clear(Me);
			D.FreePStr(FirstLine^.Data, 'Ctl - Memo_LoadFile');
			Dispose(FirstLine);
			FirstLine:=FirstBlockLine;
			FirstLine^.Prev:=Nil;
			CursorLine:=1;
			CursorChar:=0;
			CursorLinePtr:=FirstLine
		End
		Else Begin { ⠢ i 䠩  i  }
			TempStr:=String(CursorLinePtr^.Data^);
			D.ReassignPStr(Copy(TempStr, 1, CursorChar), CursorLinePtr^.Data);
			NextLine:=CursorLinePtr^.Next;
			CursorLinePtr^.Next:=FirstBlockLine;
			FirstBlockLine^.Prev:=CursorLinePtr;
			If NextLine<>Nil Then
			Begin
				New(CurLine^.Next);
				CurLine^.Next^.Prev:=CurLine;
				D.AssignPStr(Copy(TempStr, CursorChar+1, Length(TempStr)-CursorChar),
					CurLine^.Next^.Data);
				CurLine^.Next^.Kind:=CursorLinePtr^.Kind;
				CurLine^.Next^.Next:=NextLine;
				NextLine^.Prev:=CurLine^.Next;
				If Not LastLineParaEnd Then
					CurLine^.Kind:=tlNormal {?}
			End
			Else
				CurLine^.Next:=Nil;
			CursorLinePtr^.Kind:=tlNormal;
			CursorLinePtr:=CurLine;
			CursorLine:=Memo_IndexByLine(Me, CursorLinePtr);
			CursorChar:=Length(String(CursorLinePtr^.Data^));

			Memo_Reformat(Me);
			If CursorLine>FirstVisLine+VisLen-1 Then
				Memo_Scroll(Me, CursorLine-(FirstVisLine+VisLen-1), True)
		End;

		SelPresent:=False;
		Memo_CountLines(Me);
		Memo_TryAutoRedraw(Me)
	End
End;

Procedure Memo_SaveToFile(Me:PControl; FName:String;
	KeepWordWrap,SaveBAK,DoRecodeBack:Boolean);
Var
	F: Text;
	CurLine: PTextLine;
	TempStr: String;
	V: Byte;
Begin
	With Me^, TMemo(Data^) Do
	Begin

		If SaveBak And D.FileExist(FName) Then
		Begin
			If D.FileExist(D.ClipFPath(FName)+D.ClipFName(FName)+'.bak') Then
			Begin
				Assign(F, D.ClipFPath(FName)+D.ClipFName(FName)+'.bak');
				Erase(F)
			End;
			Assign(F, FName);
			Rename(F, D.ClipFPath(FName)+D.ClipFName(FName)+'.bak')
		End;
		Assign(F, FName);
		Rewrite(F);
		CurLine:=FirstLine;
		While CurLine<>Nil Do
		Begin
			If DoRecodeBack And (Coding<>0) Then
			Begin
				TempStr:=String(CurLine^.Data^);
				For V:=1 To Length(TempStr) Do
					TempStr[V]:=RecodeBack[Coding, Byte(TempStr[V])];
				Write(F, TempStr)
			End
			Else
				Write(F, String(CurLine^.Data^));

			If (KeepWordWrap Or (CurLine^.Kind=tlParaEnd)) And
				(CurLine^.Next<>Nil) Then
				WriteLn(F);
			CurLine:=CurLine^.Next
		End;
		Close(F)
	End
End;

Procedure Memo_SaveSelToFile(Me:PControl; FName:String);
Var
	F: Text;
	CurLine: PTextLine;
	CurLineNum: Integer;
	Temp:String;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		If Not SelPresent Then Exit;
		Assign(F, FName);
		Rewrite(F);
		If SelStartLine=SelEndLine Then
		Begin { i    浪 }
			CurLine:=Memo_LineByIndex(Me, SelStartLine);
			Temp:=String(CurLine^.Data^);
			Temp:=Copy(Temp, SelStartChar+1, SelEndChar-SelStartChar);
			Write(F, Temp)
		End
		Else Begin { 浪 i }
				{   䠩 訩 (⪮) 冷 }
			CurLineNum:=SelStartLine;
			CurLine:=Memo_LineByIndex(Me, SelStartLine);
			Temp:=String(CurLine^.Data^);
			Temp:=Copy(Temp, SelStartChar+1, Length(Temp)-SelStartChar);
				{  i 浪, i ⠭쮣 }
			Repeat
				Write(F, Temp);
				If CurLine^.Kind=tlParaEnd Then WriteLn(F);
				CurLine:=CurLine^.Next;
				Inc(CurLineNum);
				Temp:=String(CurLine^.Data^);
			Until CurLineNum=SelEndLine;
				{  ⠭i (⪮) 冷 }
			Temp:=Copy(Temp, 1, SelEndChar);
			Write(F, Temp)
		End;
		Close(F)
	End
End;





Procedure Memo_Select(Me:PControl; LastLine, LastChar:Integer);
Begin
	With Me^, TMemo(Data^) Do
	Begin
		{if (not sel) or ((sel1<>cursor) and (sel2<>cursor))}
		If (Not SelPresent) Or
			((LastLine<>SelStartLine) And (LastChar<>SelStartChar) And
			(LastLine<>SelEndLine) And (LastChar<>SelEndChar)) Then
		Begin
			SelStartLine:=CursorLine;
			SelStartChar:=CursorChar;
			SelEndLine:=LastLine;
			SelEndChar:=LastChar;
			SelPresent:=True
		End
		Else Begin
			If (LastLine=SelStartLine) And (LastChar=SelStartChar) Then
			Begin
				SelStartLine:=CursorLine;
				SelStartChar:=CursorChar
			End
			Else Begin
				SelEndLine:=CursorLine;
				SelEndChar:=CursorChar
			End;
			If (SelStartLine=SelEndLine) And (SelStartChar=SelEndChar) Then
				SelPresent:=False;
		End;
		Case D.Sgn(SelEndLine-SelStartLine) Of
			-1:Begin
				D.SwapInt(SelStartLine, SelEndLine);
				D.SwapInt(SelStartChar, SelEndChar)
			End;
			0:If SelStartChar>SelEndChar Then
				D.SwapInt(SelStartChar, SelEndChar);
		End;
		If (SelStartLine=SelEndLine) And (SelStartChar=SelEndChar) Then
			SelPresent:=False;
		If LastLine<=CursorLine
			Then Memo_ShowLines(Me, LastLine, CursorLine-LastLine+1, True)
			Else Memo_ShowLines(Me, CursorLine, LastLine-CursorLine+1, True)
	End
End;

Procedure Memo_Unselect(Me:PControl);
Begin
	With Me^, TMemo(Data^) Do
	Begin
		If SelPresent Then
		Begin
			SelPresent:=False;
			Memo_ShowLines(Me, SelStartLine, SelEndLine-SelStartLine+1, True);
			SelStartLine:=0;
			SelEndLine:=0
		End
	End
End;

Procedure Memo_DelSel(Me:PControl);
Var
	FirstSelLine, CurLine, NextLine: PTextLine;
	CurLineNum: Integer;
	Temp:String;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		Log.BugMsg('Memo', 'DelSel', 'Starting...');
		If Not SelPresent Then Exit;
		SelPresent:=False;
		If SelStartLine=SelEndLine Then
		Begin {  i쪠 i ।i 浪 }
			Log.BugMsg('Memo', 'DelSel', 'Deleting one line fragment');
			Temp:=Copy(String(CursorLinePtr^.Data^), 1, SelStartChar);
			Temp:=Temp+Copy(String(CursorLinePtr^.Data^), SelEndChar+1,
				Length(String(CursorLinePtr^.Data^))-SelEndChar);
			D.ReassignPStr(Temp, CursorLinePtr^.Data);
			CursorChar:=SelStartChar;
			Memo_ReformatParaEnd(Me, CursorLinePtr);
			Memo_RefreshCursorX(Me);
			If String(CursorLinePtr^.Data^)=Temp
				Then Memo_ShowLines(Me, CursorLine, 1, True)
				Else Memo_IntDraw(Me, True)
		End
		Else Begin {  i쪠 浪i. i ᪫ . }
			Log.BugMsg('Memo', 'DelSel', 'Deleting multiline block...');
				{ 01.11.2003 ४,      }
			If (CursorLine=SelEndLine) And (CursorChar=SelEndChar) Then
			Begin
				CursorLine:=SelStartLine;
				CursorChar:=SelStartChar
			End;
			FirstSelLine:=Memo_LineByIndex(Me, SelStartLine);
			CurLine:=FirstSelLine^.Next;
			CurLineNum:=SelStartLine+1;
				{ 1.  i 浪 i 訬 i ⠭i,    }
			While CurLineNum<SelEndLine Do
			Begin
				D.FreePStr(CurLine^.Data, 'Memo - DelSel');
				NextLine:=CurLine^.Next;
				Dispose(CurLine);
				CurLine:=NextLine;
				Inc(CurLineNum)
			End;
			Log.BugMsg('Memo', 'DelSel', 'Deleted lines between first and last');
				{ 2.  i 讣 i ⮪ ⠭쮣 浪 }
			Temp:=Copy(String(FirstSelLine^.Data^), 1, SelStartChar);
			D.ReassignPStr(Temp, FirstSelLine^.Data);
			Temp:=Copy(String(CurLine^.Data^), SelEndChar+1,
				Length(String(CurLine^.Data^))-SelEndChar);
			Log.BugMsg('Memo', 'DelSel', 'Last line = "'+Temp+'"');
			D.ReassignPStr(Temp, CurLine^.Data);
			Log.BugMsg('Memo', 'DelSel', 'Dropped deleted text in first and last lines');
				{ 3. ࠢ ᨫ 讣 i ⠭쮣 浪 i ⨯ }
			FirstSelLine^.Next:=CurLine;
			CurLine^.Prev:=FirstLine;
			FirstSelLine^.Kind:=tlNormal;
			Log.BugMsg('Memo', 'DelSel', 'Links updated');
				{ 4. i ⠭, i ⨭ ஡ }
				{ (󬭮 ४ ᭮ ஡  i :) }
			Memo_ReformatParaEnd(Me, FirstSelLine);
			Log.BugMsg('Memo', 'DelSel', 'Paragraph end reformatted');
				{ ଠi - ६ i ࠢ  }
			CursorChar:=SelStartChar;
			CursorLine:=SelStartLine;
			CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);
			Memo_RefreshCursorX(Me);
			Memo_IntDraw(Me, True);
			Log.BugMsg('Memo', 'DelSel', 'Almost done...');
		End;
		Memo_CountLines(Me);
		Log.BugMsg('Memo', 'DelSel', 'Counted lines. Now finished!');
	End
End;





Function Memo_CharByCoord(Me:PControl; S:String; Coord:Integer):Integer;
Var
	BestPos:Byte;
	V, CurPos, BestDist:Integer;
Begin
	With Me^, Place, TMemo(Data^) Do
	Begin
		{     浪 }
		Font.SetFont(MyFont);
		CurPos:=X1+3;
		BestPos:=0;
		BestDist:=Abs(CurPos-Coord);
		For V:=1 To Length(S) Do
		Begin
			Inc(CurPos, Font.Width(S[V]));
			If Abs(CurPos-Coord)<BestDist Then
			Begin
				BestPos:=V;
				BestDist:=Abs(CurPos-Coord)
			End
		End
	End;
	Memo_CharByCoord:=BestPos
End;

Procedure Memo_ScrollTimer(Me:PControl;Times:Integer);
Var
	LastLine, LastChar: Integer;
Begin
	With Me^, Place, TMemo(Data^) Do
	Begin
		LastLine:=CursorLine;
		LastChar:=CursorChar;
		If D.InRectCoord(Kernel.MsX, Kernel.MsY, X1, Y1, X2, Y1+3) And
			(FirstVisLine>1) Then
		Begin
			Memo_Scroll(Me, -1, True);
			Kernel.BeginPaint(Place);
			Dec(CursorLine);
			CursorChar:=0;
			CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);
			Memo_RefreshCursorX(Me);
			Memo_Select(Me, LastLine, LastChar);
			Kernel.EndPaint
		End;
		If D.InRectCoord(Kernel.MsX, Kernel.MsY, X1, Y2-3, X2, Y2) And
			(FirstVisLine+VisLen-1<TotalLines) Then
		Begin
			Memo_Scroll(Me, 1, True);
			Kernel.BeginPaint(Place);
			Inc(CursorLine);
			CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);
			CursorChar:=Length(String(CursorLinePtr^.Data^));
			Memo_RefreshCursorX(Me);
			Memo_Select(Me, LastLine, LastChar);
			Kernel.EndPaint
		End;
	End
End;

Procedure Memo_HideMouse(Var X,Y:Integer);
Begin
	LastX:=X;
	LastY:=Y; { vars from InpLine.Inc }
	Kernel.HideMouse(X,Y)
End;

Procedure Memo_ShowMouse(Var X,Y:Integer);
Var
	LastLine, LastChar: Integer;
Begin
	With CurCtl^, Place, TMemo(Data^) Do
	Begin
		D.ToRange(X, X1, X2);
		D.ToRange(Y, Y1, Y2);
		If D.InRectCoord(X, Y, X1, Y1+3, X2, Y1+3+VisLen*MyFontHeight) Then
		Begin
			LastLine:=CursorLine;
			LastChar:=CursorChar;
			CursorLine:=(Y-Y1-4) Div MyFontHeight+FirstVisLine;
			D.ToRange(CursorLine, 1, TotalLines);
			CursorLinePtr:=Memo_LineByIndex(CurCtl, CursorLine);
			CursorChar:=Memo_CharByCoord(CurCtl, String(CursorLinePtr^.Data^), X);
			If (LastLine<>CursorLine) Or (LastChar<>CursorChar) Then
			Begin
				Memo_RefreshCursorX(CurCtl);
				Memo_Select(CurCtl, LastLine, LastChar)
			End
		End
	End;
	Kernel.ShowMouse(X,Y)
End;

Procedure Memo_EditCut(Me:PControl);
Begin
	Memo_SaveSelToFile(Me, 'clipbrd.txt');
	Memo_DelSel(Me);
	If D.HostOS=osWindows Then
		D.WinCopy
End;

Procedure Memo_EditCopy(Me:PControl);
Begin
	Memo_SaveSelToFile(Me, 'clipbrd.txt');
	If D.HostOS=osWindows Then
		D.WinCopy
End;

Procedure Memo_EditPaste(Me:PControl);
Begin
	If D.HostOS=osWindows Then
		D.WinPaste;
	If D.FileExist('clipbrd.txt') Then
	Begin
		Memo_DelSel(Me);
		Memo_LoadFile(Me, 'clipbrd.txt', False)
	End
End;

Procedure Memo_SelectAll(Me:PControl);
Begin
	With Me^, TMemo(Data^) Do
	Begin
		SelStartLine:=1;
		SelStartChar:=0;
		SelEndLine:=TotalLines;
		SelEndChar:=Length(String(Memo_LineByIndex(Me, SelEndLine)^.Data^));
		If (SelStartLine=SelEndLine) And (SelStartChar=SelEndChar)
			Then SelPresent:=False
			Else SelPresent:=True;
		Memo_IntDraw(Me, False)
	End
End;




Procedure Memo_OnContextMenuClick(Me:PProg; Item:PMenuItem);
Begin
	If String(Item^.Name^)='i' Then
		Memo_EditCut(CurCtl);
	If String(Item^.Name^)='i' Then
		Memo_EditCopy(CurCtl);
	If String(Item^.Name^)='⠢' Then
		Memo_EditPaste(CurCtl);
	If String(Item^.Name^)='i ' Then
		Memo_SelectAll(CurCtl);
End;

Procedure Memo_OnClick(Me:PControl; X, Y:Integer);
Var
	V:Integer;
	CurLine:PTextLine;
	CurLineNum:Integer;
	Temp:String;
	Found:Boolean;
	BestPos:Byte;
	CurPos, BestDist:Integer;
Begin
	With Me^, Place, TMemo(Data^) Do
	Begin
		If Kind=mkReadOnly Then Exit;
		If Not D.InRectCoord(X, Y, X1+3, Y1+4, X2-3, Y2-4) Then Exit;
		If Mouse.GetMask=Mouse.mbRight Then
		Begin { i ࠢ  -  ⥪⭥  }
			Repeat
				Kernel.CheckSysKeys
			Until Kernel.HeartOfPolit(Kernel.HideMouse,
				Kernel.ShowMouse, Kernel.RefreshNeeded)=Mouse.mbNone;

			CurCtl:=Me;
			Popup.OpenMenu(Owner, ContextMenu, Kernel.MsX, Kernel.MsY);
			Exit
		End;
			{   i⮣ 浪 }
		V:=(Y-Y1-4) Div MyFontHeight;
		If Not D.InRange(FirstVisLine+V, FirstVisLine, TotalLines) Then Exit;
		Memo_ShowCursor(Me, False); { 客  }
		CursorLine:=FirstVisLine+V;
		CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);;
		CursorChar:=Memo_CharByCoord(Me, String(CursorLinePtr^.Data^), X);
		If SelPresent Then
		Begin
			SelPresent:=False;
			Memo_IntDraw(Me, False)
		End;
		Memo_RefreshCursorX(Me);
		Memo_ShowCursor(Me, True);
		Kernel.SelectCursor(Kernel.MouseText);
		CurCtl:=Me;
		Kernel.AddTimer(Me, Memo_ScrollTimer, 3);

		Repeat
			Kernel.CheckSysKeys
		Until Kernel.HeartOfPolit(Memo_HideMouse, Memo_ShowMouse,
			Kernel.RefreshNeeded)=0;

		Kernel.RemoveTimer(Me, Memo_ScrollTimer);
		Kernel.SelectCursor(Kernel.MouseDefault)
	End
End;

Procedure Memo_KeyEnter(Me:PControl);
Var
	NextLine:PTextLine;
	TempStr:String;
Begin
	With Me^, Place, TMemo(Data^) Do
	Begin
		TempStr:=String(CursorLinePtr^.Data^);
		NextLine:=CursorLinePtr^.Next;
		New(CursorLinePtr^.Next);
		CursorLinePtr^.Next^.Prev:=CursorLinePtr;
		CursorLinePtr^.Next^.Next:=NextLine;
		CursorLinePtr^.Next^.Kind:=CursorLinePtr^.Kind;
		CursorLinePtr^.Kind:=tlParaEnd;
		If NextLine<>Nil Then
			NextLine^.Prev:=CursorLinePtr^.Next;
		D.ReassignPStr(Copy(TempStr, 1, CursorChar), CursorLinePtr^.Data);
		TempStr:=Copy(TempStr, CursorChar+1, Length(TempStr)-CursorChar);
		D.AssignPStr(TempStr, CursorLinePtr^.Next^.Data);

		Inc(CursorLine);
		CursorChar:=0;
		CursorLinePtr:=CursorLinePtr^.Next;
		Memo_RefreshCursorX(Me);

		Memo_ReformatParaEnd(Me, CursorLinePtr);
		If CursorLine>FirstVisLine+VisLen-1
			Then Memo_Scroll(Me, 1, True)
			Else Memo_IntDraw(Me, True)
	End
End;

Procedure Memo_KeyDel(Me:PControl);
Var
	TempStr:String;
	CurLine, NextLine:PTextLine;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		Memo_ShowCursor(Me, False);
		TempStr:=String(CursorLinePtr^.Data^);
		If CursorChar<Length(TempStr) Then
		Begin {  i ।i 浪 }
			TempStr:=Copy(TempStr, 1, CursorChar)+
				Copy(TempStr, CursorChar+2, Length(TempStr)-CursorChar-1);
			D.ReassignPStr(TempStr, CursorLinePtr^.Data);
			If CursorLinePtr^.Prev<>Nil
				Then If CursorLinePtr^.Prev^.Kind<>tlParaEnd
					Then Memo_ReformatParaEnd(Me, CursorLinePtr^.Prev)
					Else Memo_ReformatParaEnd(Me, CursorLinePtr)
				Else Memo_ReformatParaEnd(Me, CursorLinePtr);
			If TempStr=String(CursorLinePtr^.Data^)
				Then Memo_ShowLines(Me, CursorLine, 1, True)
				Else Memo_IntDraw(Me, True)
		End
		Else Begin
			If CursorLinePtr^.Next<>Nil Then {  i ⥪? }
			Begin { `󤭠   }
				TempStr:=String(CursorLinePtr^.Data^)+
					String(CursorLinePtr^.Next^.Data^);
				D.ReassignPStr(TempStr, CursorLinePtr^.Data);
				CursorLinePtr^.Kind:=CursorLinePtr^.Next^.Kind;
				NextLine:=CursorLinePtr^.Next^.Next;
				D.FreePStr(CursorLinePtr^.Next^.Data, 'Ctl - Memo_KeyDel');
				Dispose(CursorLinePtr^.Next);
				CursorLinePtr^.Next:=NextLine;
				If NextLine<>Nil Then
					NextLine^.Prev:=CursorLinePtr;
				Memo_ReformatParaEnd(Me, CursorLinePtr);
				Memo_IntDraw(Me, True)
			End
		End;
		Memo_ShowCursor(Me, True)
	End
End;

Procedure Memo_KeyLeft(Me:PControl);
Var
	LastLine, LastChar: Integer;
Begin
	With Me^, Place, TMemo(Data^) Do
	Begin
		LastLine:=CursorLine;
		LastChar:=CursorChar;
		Memo_ShowCursor(Me, False);
		If CursorChar=0 Then
		Begin
			If CursorLine>1 Then
			Begin
				Dec(CursorLine);
				If CursorLine<FirstVisLine Then
					Memo_Scroll(Me, -1, True);
				{CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);}
				CursorLinePtr:=CursorLinePtr^.Prev;
				CursorChar:=Length(String(CursorLinePtr^.Data^))
			End
		End
		Else
			Dec(CursorChar);

		If (D.KeyStatus And (ksLeftShift+ksRightShift))<>0
			Then Memo_Select(Me, LastLine, LastChar)
			Else Memo_Unselect(Me);
		Memo_RefreshCursorX(Me);
		Memo_ShowCursor(Me, True)
	End
End;

Function Memo_OnKey(Me:PControl; Key, ExtKey:Char):Boolean;
Var
	Recognized:Boolean;
	LastFirstLine:Integer;
	TempStr:String;
	LastLine, LastChar: Integer;
Begin
	Recognized:=False;
	With Me^, Place, TMemo(Data^) Do
	Begin
		LastFirstLine:=FirstVisLine;

		If Kind=mkReadOnly Then
		Begin
			If (Key=#0) And (ExtKey=#72) Then
			Begin { Up arrow }
				If FirstVisLine>1 Then
					Memo_Scroll(Me, -1, True);
				Recognized:=True
			End;
			If (Key=#0) And (ExtKey=#80) Then
			Begin { Down arrow }
				If (FirstVisLine+VisLen-1<TotalLines) Then
					Memo_Scroll(Me, 1, True);
				Recognized:=True
			End;
			If (Key=#0) And (ExtKey=#73) Then
			Begin { PgUp }
				FirstVisLine:=D.Max(1, FirstVisLine-VisLen+1);
				If FirstVisLine<>LastFirstLine Then
					Memo_IntDraw(Me, True);
				Recognized:=True
			End;
			If (Key=#0) And (ExtKey=#81) Then
			Begin { PgDn }
				FirstVisLine:=D.Min(FirstVisLine+VisLen-1, TotalLines-VisLen+1);
				If FirstVisLine<>LastFirstLine Then
					Memo_IntDraw(Me, True);
				Recognized:=True
			End;
			If (Key=#0) And (ExtKey=#71) Then
			Begin { Home }
				FirstVisLine:=1;
				If FirstVisLine<>LastFirstLine Then
					Memo_IntDraw(Me, True);
				Recognized:=True
			End;
			If (Key=#0) And (ExtKey=#79) Then
			Begin { End }
				FirstVisLine:=TotalLines-VisLen+1;
				If FirstVisLine<>LastFirstLine Then
					Memo_IntDraw(Me, True);
				Recognized:=True
			End
		End


		Else Begin { Kind=mkEditable }
			Font.SetFont(MyFont);
			LastLine:=CursorLine;
			LastChar:=CursorChar;
				{   ࠧ 冷 i ஬? }
			If FirstVisLine>CursorLine Then
			Begin
				FirstVisLine:=CursorLine;
				Memo_IntDraw(Me, True)
			End;
			If FirstVisLine+VisLen-1<CursorLine Then
			Begin
				FirstVisLine:=CursorLine-VisLen+1;
				Memo_IntDraw(Me, True)
			End;

			If (Key=#0) And (ExtKey=#75) Then
			Begin { Left arrow }
				Memo_KeyLeft(Me);
				Recognized:=True
			End;
			If (Key=#0) And (ExtKey=#77) Then
			Begin { Right arrow }
				Memo_ShowCursor(Me, False);
				If CursorChar=Length(String(CursorLinePtr^.Data^)) Then
				Begin
					If CursorLine<TotalLines Then
					Begin
						Inc(CursorLine);
						If CursorLine>FirstVisLine+VisLen-1 Then
							Memo_Scroll(Me, 1, True);
						CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);
						CursorChar:=0
					End;
				End
				Else
					Inc(CursorChar);

				If (D.KeyStatus And (ksLeftShift+ksRightShift))<>0
					Then Memo_Select(Me, LastLine, LastChar)
					Else Memo_Unselect(Me);
				Memo_RefreshCursorX(Me);
				Memo_ShowCursor(Me, True);
				Recognized:=True
			End;

			If (Key=#0) And (ExtKey=#72) Then
			Begin { Up arrow }
				Memo_ShowCursor(Me, False);
				D.DecNoLess(CursorLine, 1);
				If CursorLine<FirstVisLine Then
					Memo_Scroll(Me, -1, True);
				CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);
				CursorChar:=Memo_CharByCoord(Me, String(CursorLinePtr^.Data^), CursorX);
				If (D.KeyStatus And (ksLeftShift+ksRightShift))<>0
					Then Memo_Select(Me, LastLine, LastChar)
					Else Memo_Unselect(Me);
				Memo_RefreshCursorX(Me);
				Memo_ShowCursor(Me, True);
				Recognized:=True
			End;
			If (Key=#0) And (ExtKey=#80) Then
			Begin { Down arrow }
				Memo_ShowCursor(Me, False);
				D.IncNoMore(CursorLine, TotalLines);
				If CursorLine>FirstVisLine+VisLen-1 Then
					Memo_Scroll(Me, 1, True);
				CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);
				CursorChar:=Memo_CharByCoord(Me, String(CursorLinePtr^.Data^), CursorX);
				If (D.KeyStatus And (ksLeftShift+ksRightShift))<>0
					Then Memo_Select(Me, LastLine, LastChar)
					Else Memo_Unselect(Me);
				Memo_RefreshCursorX(Me);
				Memo_ShowCursor(Me, True);
				Recognized:=True
			End;

			If (Key=#0) And (ExtKey=#71) Then
			Begin { Home }
				Memo_ShowCursor(Me, False);
				CursorChar:=0;
				If (D.KeyStatus And (ksLeftShift+ksRightShift))<>0
					Then Memo_Select(Me, LastLine, LastChar)
					Else Memo_Unselect(Me);
				Memo_RefreshCursorX(Me);
				Memo_ShowCursor(Me, True);
				Recognized:=True
			End;
			If (Key=#0) And (ExtKey=#79) Then
			Begin { End }
				Memo_ShowCursor(Me, False);
				CursorChar:=Length(String(CursorLinePtr^.Data^));
				If (D.KeyStatus And (ksLeftShift+ksRightShift))<>0
					Then Memo_Select(Me, LastLine, LastChar)
					Else Memo_Unselect(Me);
				Memo_RefreshCursorX(Me);
				Memo_ShowCursor(Me, True);
				Recognized:=True
			End;

			If Key In [' '..#255] Then
			Begin { alphabetical/numeric char }
				If SelPresent Then
					Memo_DelSel(Me);
				TempStr:=String(CursorLinePtr^.Data^);
				TempStr:=Copy(TempStr, 1, CursorChar)+Key+
					Copy(TempStr, CursorChar+1, Length(TempStr)-CursorChar);
				D.ReassignPStr(TempStr, CursorLinePtr^.Data);
				Memo_ShowCursor(Me, False);
				Inc(CursorChar);
				Memo_RefreshCursorX(Me);
				Font.SetFont(MyFont);

				If (Font.Width(TempStr)>X2-X1-8) Or (Key=' ') Then
				Begin
					If Key=' ' Then
						If CursorLinePtr^.Prev<>Nil
							Then If CursorLinePtr^.Prev^.Kind<>tlParaEnd
								Then Memo_ReformatParaEnd(Me, CursorLinePtr^.Prev)
								Else Memo_ReformatParaEnd(Me, CursorLinePtr)
							Else Memo_ReformatParaEnd(Me, CursorLinePtr)
						Else Memo_ReformatParaEnd(Me, CursorLinePtr);
					If String(CursorLinePtr^.Data^)=TempStr
						Then Memo_ShowLines(Me, CursorLine, 1, True)
						Else Memo_IntDraw(Me, True)
				End
				Else
					Memo_ShowLines(Me, CursorLine, 1, True);

				If CursorLine<FirstVisLine Then
					Memo_Scroll(Me, CursorLine-FirstVisLine, True);
				If CursorLine>FirstVisLine+VisLen-1 Then
					Memo_Scroll(Me, CursorLine-(FirstVisLine+VisLen-1), True);
				Memo_ShowCursor(Me, True);
				Recognized:=True
			End;

			If (Key=#0) And (ExtKey=#83) Then
			Begin { del }
				If SelPresent
					Then Memo_DelSel(Me)
					Else Memo_KeyDel(Me);
				Recognized:=True
			End;
			If Key=#13 Then
			Begin { Enter }
				Memo_KeyEnter(Me);
				Recognized:=True
			End;
			If (Key=#8) And Not ((CursorLine=1) And (CursorChar=0)) Then
			Begin { Backspace }
				If Not SelPresent Then
				Begin
					Memo_KeyLeft(Me);
					Memo_KeyDel(Me)
				End
				Else
					Memo_DelSel(Me);
				Recognized:=True
			End;

			If (Key=#0) And (ExtKey=#73) Then
			Begin { PgUp }
				If CursorLine>1 Then
				Begin
					Memo_ShowCursor(Me, False);
					If CursorLine=FirstVisLine Then
					Begin
						CursorLine:=D.Max(1, FirstVisLine-VisLen+1);
						FirstVisLine:=CursorLine;
						Memo_IntDraw(Me, True)
					End
					Else
						CursorLine:=FirstVisLine;
					CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);
					CursorChar:=Memo_CharByCoord(Me, String(CursorLinePtr^.Data^), CursorX);
					If (D.KeyStatus And (ksLeftShift+ksRightShift))<>0
						Then Memo_Select(Me, LastLine, LastChar)
						Else Memo_Unselect(Me);
					Memo_RefreshCursorX(Me);
					Memo_ShowCursor(Me, True);
				End;
				Recognized:=True
			End;
			If (Key=#0) And (ExtKey=#81) Then
			Begin { PgDn }
				If CursorLine<TotalLines Then
				Begin
					Memo_ShowCursor(Me, False);
					If CursorLine=FirstVisLine+VisLen-1 Then
					Begin
						CursorLine:=D.Min(TotalLines, CursorLine+VisLen-1);
						FirstVisLine:=D.Max(1, CursorLine-VisLen+1);
						Memo_IntDraw(Me, True)
					End
					Else
						CursorLine:=D.Min(FirstVisLine+VisLen-1, TotalLines);
					CursorLinePtr:=Memo_LineByIndex(Me, CursorLine);
					CursorChar:=Memo_CharByCoord(Me, String(CursorLinePtr^.Data^), CursorX);
					If (D.KeyStatus And (ksLeftShift+ksRightShift))<>0
						Then Memo_Select(Me, LastLine, LastChar)
						Else Memo_Unselect(Me);
					Memo_RefreshCursorX(Me);
					Memo_ShowCursor(Me, True);
				End;
				Recognized:=True
			End;

			If Key=^X Then
			Begin
				Memo_EditCut(Me);
				Recognized:=True
			End;
			If Key=^C Then
			Begin
				Memo_EditCopy(Me);
				Recognized:=True
			End;
			If Key=^V Then
			Begin
				Memo_EditPaste(Me);
				Recognized:=True
			End;
			If Key=^A Then
			Begin
				Memo_SelectAll(Me);
				Recognized:=True
			End;

			If (Key=#0) And (ExtKey=#60) Then
			Begin { F2 - temp save }
				Memo_SaveToFile(Me, 'text.txt', False, True, True);
				Recognized:=True
			End;

		End
	End;
	Memo_OnKey:=Recognized
End;

Procedure Memo_OnRefocus(Me:PControl; Focused:Boolean);
Begin
	Inc(Me^.Place.X2, 16);
	Kernel.CtlRefocusDummy(Me, Focused);
	Dec(Me^.Place.X2, 16);
	If TMemo(Me^.Data^).Kind=mkEditable Then
		Memo_ShowCursor(Me, Focused)
End;

Procedure Memo_OnScrollUpdate(Me:PProg; Control:PControl);
Var
	LastFirstLine: Integer;
Begin
	With TMemo(Control^.Data^) Do
		Memo_Scroll(Control, Scroll_GetPos(MemoScroll)-FirstVisLine, False)
End;

Procedure Memo_CursorTimer(Me:PControl; Times:Integer);
Var
	CurY: Integer;
Begin
	With Me^, Place, TMemo(Data^) Do
	Begin
		If Kind=mkReadOnly Then Exit;
		If (Not CanDraw) Or (Me^.Owner<>Kernel.ActiveWin) Or
			(Not Visible) Or (Me<>Me^.Owner^.ActiveCtl) Then Exit;

		If Not D.InRange(CursorLine, FirstVisLine, FirstVisLine+VisLen-1) Then
			Exit;

		If CursorColor=ColorFont
			Then Memo_ShowCursor(Me, False)
			Else Memo_ShowCursor(Me, True)
	End
End;

Procedure Memo_TextCursorTimer(Me:PControl; Times:Integer);
Var
	CurY: Integer;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		If Kind=mkReadOnly Then Exit;
		If (Not CanDraw) Or (Not Visible) Then Exit;
		If D.InRect(Kernel.MsX, Kernel.MsY, Place) Then
			Kernel.SelectCursor(Kernel.MouseText)
	End
End;

Procedure Memo_OnWheel(Me:PControl; Delta:Integer);
Begin
	Memo_Scroll(Me, Delta*3, True)
End;

Procedure Memo_Print(Me:PControl);
Var
	CurLine: PTextLine;
Begin
	With Me^, TMemo(Data^) Do
	Begin
		CurLine:=FirstLine;
		While CurLine<>Nil Do
		Begin
			{WriteLn(Lst, String(CurLine^.Data^));}
			LptGfx.PrintStr(String(CurLine^.Data^));
			CurLine:=CurLine^.Next
		End;
		{Write(Lst, #12)} { page feed ? }
	End
End;




Procedure Memo_OnDestroy(Me:PControl);
Begin
	Memo_Clear(Me);
	With TMemo(Me^.Data^) Do
	Begin
		D.FreePStr(FirstLine^.Data, 'Ctl - Memo_OnDestroy');
		Dispose(FirstLine);
		Popup.FreeMenu(ContextMenu)
	End;
	Kernel.RemoveTimer(Me, Memo_CursorTimer);
	Kernel.RemoveTimer(Me, Memo_TextCursorTimer);
	FreeMem(Me^.Data, SizeOf(TMemo))
End;

Procedure Memo_SetKind(Me:PControl; VKind:TMemoKind);
Begin
	With Me^, TMemo(Data^) Do
	Begin
		Kind:=VKind;
		Case Kind Of
			mkReadOnly:Begin
				ColorBG:=Kernel.ClrDisabled;
				ColorFont:=Kernel.ClrTextCom;
				SelPresent:=False
			End;
			mkEditable:Begin
				ColorBG:=Kernel.ClrEdit;
				ColorFont:=Kernel.ClrEditFont
			End
		End;

		Memo_TryAutoRedraw(Me)
	End
End;

Procedure Memo_SetFont(Me:PControl; VFont:Pointer);
Begin
	With Me^, TMemo(Data^) Do
	Begin
		MyFont:=VFont;
		MyFontHeight:=Font.Height(MyFont);
	End;
	Memo_Reformat(Me);
	Memo_TryAutoRedraw(Me)
End;

Function Memo_Create(Prog:PProg; VKind:TMemoKind; Enabled:Boolean; Hint:String):PControl;
Var
	Me:PControl;
	TempPeriod:Integer;
Begin
	Me:=Kernel.AddCtl(Prog);
		{Memo_OnClick, Memo_OnKey,
		Memo_OnRefocus, Kernel.CtlDummy, Memo_OnDestroy,
		Enabled, True, Hint);}

	With Me^ Do
	Begin
		OnClick:=Memo_OnClick;
		OnKey:=Memo_OnKey;
		OnRefocus:=Memo_OnRefocus;
		OnWheel:=Memo_OnWheel;
		OnDestroy:=Memo_OnDestroy;

		GetMem(Data, SizeOf(TMemo));
		With TMemo(Data^) Do
		Begin
			LastWidth:=0; { minimum }
			FirstLine:=Nil;
			Memo_Clear(Me);
			Memo_SetKind(Me, VKind);
			SelPresent:=False;
			SelStartLine:=0;
			SelEndLine:=0;
			MyFont:=Kernel.CurFont;
			MyFontHeight:=Font.Height(MyFont);
			Coding:=0;

			MemoScroll:=Scroll_Create(Owner, 0, 0, 0, orVertical,
				Memo_OnScrollUpdate);
			MemoScroll^.FocusCtl:=Me;
			MemoScroll^.OwnerCtl:=Me;
			MemoScroll^.AutoRedraw:=False;

			ContextMenu:=Popup.CreateMenu;
			Popup.AddCommand(ContextMenu, 'i', 'Ctrl+X',
				'७ i ⥪  ',
				Kernel.PicCut, True, Memo_OnContextMenuClick);
			Popup.AddCommand(ContextMenu, 'i', 'Ctrl+C',
				'i i ⥪  ',
				Kernel.PicCopy, True, Memo_OnContextMenuClick);
			Popup.AddCommand(ContextMenu, '⠢', 'Ctrl+V',
				'⠢ ⥪ i ',
				Kernel.PicPaste, True, Memo_OnContextMenuClick);
			Popup.AddSpace(ContextMenu);
			Popup.AddCommand(ContextMenu, 'i ', 'Ctrl+A',
				'i  ⥪  ।i',
				Nil, True, Memo_OnContextMenuClick)
		End
	End;
	Ini.Select('polit.ini', 'Controls');
	TempPeriod:=Ini.GetInt('CursorBlinkPeriod', 5);
	If TempPeriod>0 Then
		Kernel.AddTimer(Me, Memo_CursorTimer, TempPeriod);
	AlwaysTextCursor:=Ini.GetBool('AlwaysTextCursor', False);
	If AlwaysTextCursor Then
		Kernel.AddTimer(Me, Memo_TextCursorTimer, 3);
	Memo_Create:=Me
End;
