{ ************ Progress bar ************ }
Const
	ProgressDist=5;

Type
	TProgress=Record
		Min, Max, Value: Integer;
		Kind: TProgressKind;
	End;

Function Progress_CoordX(Me:PControl; V:Integer): Integer;
Var
	Temp: LongInt;
Begin
	With Me^, Place, TProgress(Data^) Do
		Temp:=X1+3+LongInt(X2-X1-6)*LongInt(V-Min) Div (Max-Min);
	Progress_CoordX:=Temp
End;

Procedure Progress_IntDraw(Me:PControl);
Var
	I, AX, AY: Integer;
Begin
	If ((Not Me^.CanDraw) And (Not Kernel.DontBother)) Or Not Me^.Visible Then
		Exit;
	With Me^, TProgress(Data^), Place Do
	Begin
		Gfx.Rectangle(X1, Y1, X2, Y2, Kernel.ClrTextCom);
		Gfx.Rectangle(X1+1, Y1+1, X2-1, Y2-1, Kernel.ClrWinBG);
		If (Kind=pkDeterm) And (Value=Max)
			Then Gfx.Bar(X1+2, Y1+2, X2-2, Y2-2, Kernel.Clr3DAct)
			Else Gfx.Bar(X1+2, Y1+2, X2-2, Y2-2, Kernel.ClrTextDis);
		If Kind=pkDeterm Then
		Begin
			If Value<>Min Then
				For I:=X1+3 To Progress_CoordX(Me, Value) Do
					If Odd(I-X1) Then
						Gfx.Line(I, Y1+3, I, Y2-3, Kernel.ClrTextCom)
		End
		Else
			For AX:=X1+3 To X2-3 Do
				For AY:=Y1+3 To Y2-3 Do
					If (AX+AY+Value) Mod ProgressDist=0 Then
						Gfx.PutPixel(AX, AY, Kernel.ClrTextCom)
	End
End;

Procedure Progress_Paint(Me:PControl; VX1,VY1,VX2:Integer);
Begin
	With Me^.Owner^.ClientRect Do
		D.AssignRect(X1+VX1, Y1+VY1, X1+VX2, Y1+VY1+21, Me^.Place);
	Progress_IntDraw(Me)
End;


Procedure Progress_SetValue(Me:PControl; VValue:Integer);
Var
	LastValue, I: Integer;
Begin
	With Me^, Place, TProgress(Data^) Do
	Begin
		LastValue:=Value;
		Value:=VValue;
		If Value>Max Then
			Value:=Min;
		If AutoRedraw And Not Kernel.IgnoreAutoRedraw And
			(Me^.CanDraw Or Kernel.DontBother) And Visible And
			(LastValue<>Value) Then
		Begin
			Kernel.BeginPaint(Place);
			If Kind=pkDeterm Then
			Begin
				If Value<Max Then
				Begin
					If LastValue<Value Then
					Begin
						For I:=Progress_CoordX(Me, LastValue) To Progress_CoordX(Me, Value) Do
							If Odd(I-X1) Then
								Gfx.Line(I, Y1+3, I, Y2-3, Kernel.ClrTextCom)
					End
					Else
						Gfx.Bar(Progress_CoordX(Me, Value), Y1+3,
							Progress_CoordX(Me, LastValue), Y2-3,
							Kernel.ClrTextDis);
					Gfx.Rectangle(X1+2, Y1+2, X2-2, Y2-2, Kernel.ClrTextDis)
				End
				Else
					Progress_IntDraw(Me)
			End
			Else
				Progress_IntDraw(Me);
			Kernel.EndPaint
		End
	End
End;

Procedure Progress_SetParams(Me:PControl; VMin,VMax,VValue:Integer);
Var
	LastValue, I: Integer;
Begin
	With Me^, Place, TProgress(Data^) Do
	Begin
		Min:=VMin;
		Max:=VMax;
		Value:=VValue;
		D.ToRange(Value, Min, Max);
		If AutoRedraw And Not Kernel.IgnoreAutoRedraw And
			(Me^.CanDraw Or Kernel.DontBother) And Visible Then
		Begin
			Kernel.BeginPaint(Place);
			Progress_IntDraw(Me);
			Kernel.EndPaint
		End
	End
End;


Procedure Progress_Step(Me:PControl);
Var
	LastValue, I: Integer;
	P: Pointer;
	Size: Word;
Begin
	With Me^, Place, TProgress(Data^) Do
		If Kind=pkIndeterm Then
		Begin
			If AutoRedraw And Not Kernel.IgnoreAutoRedraw Then
				If (Me^.CanDraw Or Kernel.DontBother) And Visible Then
				Begin
					D.Cycle(Value, Min, Max+1);
					{D.IncNoMore(Value, Max);}
					Kernel.BeginPaint(Place);
					Size:=Gfx.ImageSize(X2-X1-5, Y2-Y1-5);
					GetMem(P, Size);
					Gfx.GetImage(X1+4, Y1+3, X2-3, Y2-3, P);
					Gfx.PutImage(X1+3, Y1+3, P);
					FreeMem(P, Size);
					For I:=Y1+3 To Y2-3 Do
						If (X2-3+I+Value) Mod ProgressDist=0
							Then Gfx.PutPixel(X2-3, I, Kernel.ClrTextCom)
							Else Gfx.PutPixel(X2-3, I, Kernel.ClrTextDis);
					Kernel.EndPaint
				End
		End
		Else
			Progress_SetValue(Me, Value+1)
End;

Procedure Progress_OnTimer(Me:PControl;Times:Integer);
Var
	I:Integer;
Begin
	For I:=1 To Times Do
		Progress_Step(Me)
End;

Procedure Progress_SetKind(Me:PControl; VKind:TProgressKind);
Begin
	With Me^, TProgress(Data^) Do
	Begin
		If (Kind=pkIndeterm) And (VKind=pkDeterm) Then
			Kernel.RemoveTimer(Me, Progress_OnTimer);
		If (Kind=pkDeterm) And (VKind=pkIndeterm) Then
			Kernel.AddTimer(Me, Progress_OnTimer, 1);
		Kind:=VKind;

		If AutoRedraw And Not Kernel.IgnoreAutoRedraw Then
			If (Me^.CanDraw Or Kernel.DontBother) And Visible Then
			Begin
				Kernel.BeginPaint(Place);
				Progress_IntDraw(Me);
				Kernel.EndPaint
			End
	End
End;

Procedure Progress_Draw(Me:PControl);
Begin
	With Me^ Do
		If CanDraw Or Kernel.DontBother Then
		Begin
			Kernel.BeginPaint(Place);
			If Visible
				Then Progress_IntDraw(Me)
				Else Kernel.HideCtl(Me);
			Kernel.EndPaint
		End
End;


Procedure Progress_OnDestroy(Me:PControl);
Begin
	With Me^, TProgress(Data^) Do
		If Kind=pkIndeterm Then
			Kernel.RemoveTimer(Me, Progress_OnTimer);
	FreeMem(Me^.Data, SizeOf(TProgress))
End;


Function Progress_Create(Prog:PProg; VMin,VMax,VValue:Integer;
	VKind:TProgressKind): PControl;
Var
	Me:PControl;
Begin
	Me:=Kernel.AddCtl(Prog);
		{Kernel.CtlClickDummy, Kernel.CtlKeyDummy,
		Kernel.CtlRefocusDummy, Kernel.CtlDummy, Progress_Destroy,
		False, True, Hint);}
	With Me^ Do
	Begin
		OnDestroy:=Progress_OnDestroy;
		Enabled:=False;

		GetMem(Me^.Data, SizeOf(TProgress));
		With TProgress(Data^) Do
		Begin
			Kind:=VKind;
			If Kind=pkDeterm Then
			Begin
				Min:=VMin;
				Max:=VMax;
				If Max<=Min Then
					Max:=Min+1;
				Value:=VValue
			End
			Else Begin
				Min:=1;
				Max:=ProgressDist;
				Value:=1;
				Kernel.AddTimer(Me, Progress_OnTimer, 1)
			End
		End
	End;
	Progress_Create:=Me
End;
