{$F+}
Unit
	Salute;

Interface

Uses
	Kernel,Gfx;

Const
	Scale:LongInt=50;
	BlowAccel=12;
	ShellsNum=3;
	MinLightsPerShell=35;
	LightsNum=ShellsNum*130;

Type
	TNode=Record
		Radius:Integer;
		Alpha:Real
	End;

Var
	Shell:Array[1..ShellsNum] Of Record
		X,Y:LongInt;
		SX,SY:Integer;
		Color:Integer
	End;
	Light:Array[1..LightsNum] Of Record
		Live:Boolean;
		X,Y:LongInt;
		SX,SY:Integer;
		Color:Byte
	End;
	Count,I:LongInt;
	StartX,StartY:LongInt;
	ShellMinAccel,ShellMaxAccel,ShellMaxHorizAccel:LongInt;
	LightMaxAccel,LightVertAdd,MaxLightsPerShell:LongInt;
	F:File Of TNode;
	A:Array[1..100] Of TNode;
	FreeLights:Integer;

Implementation

Procedure DrawShell(X,Y,Color:LongInt);
Begin
	Gfx.Bar(X-1,Y-1,X+1,Y+1,Color+8);
	Gfx.Rectangle(X-2,Y-2,X+2,Y+2,Color)
End;

Procedure EraseShell(X,Y:LongInt);
Begin
	Gfx.Bar(X-2,Y-2,X+2,Y+2,clBlack)
End;

Procedure DrawLight(X,Y,Color:LongInt);
Begin
	Gfx.PutPixel(X,Y,Color);
	Gfx.PutPixel(X-1,Y,Color);
	Gfx.PutPixel(X+1,Y,Color);
	Gfx.PutPixel(X,Y-1,Color);
	Gfx.PutPixel(X,Y+1,Color)
End;

Procedure EraseLight(X,Y:LongInt);
Begin
	Gfx.PutPixel(X,Y,clBlack);
	Gfx.PutPixel(X-1,Y,clBlack);
	Gfx.PutPixel(X+1,Y,clBlack);
	Gfx.PutPixel(X,Y-1,clBlack);
	Gfx.PutPixel(X,Y+1,clBlack)
End;

Procedure InitShell(Index:LongInt);
Begin
	With Shell[Index] Do
	Begin
		X:=StartX*Scale;
		Y:=StartY*Scale;
		Color:=Random(6)+1;
		If Random(2)=0
			Then SX:=Random(ShellMaxHorizAccel)
			Else SX:=-Random(ShellMaxHorizAccel);
		SY:=-(ShellMinAccel+Random(ShellMaxAccel-ShellMinAccel))
	End
End;

Function FirstFreeLight:LongInt;
Var
	V:LongInt;
Begin
	FirstFreeLight:=0;
	If FreeLights=0 Then Exit;
	For V:=1 To LightsNum Do
		If Not Light[V].Live Then
		Begin
			FirstFreeLight:=V;
			Exit
		End
End;

Function SimilarColor(Color:LongInt):LongInt;
Begin
	Case Random(3) Of
		0:SimilarColor:=Color;
		1:SimilarColor:=Color+8;
		2:SimilarColor:=15
	End
End;

Procedure InitLight(FirstX,FirstY,ItColor,ItSX:LongInt);
Var
	Index:LongInt;
	Radius:LongInt;
	Alpha:Real;
Begin
	Index:=FirstFreeLight;
	If Index=0 Then Exit;
	Dec(FreeLights);
	With Light[Index] Do
	Begin
		Live:=True;
		X:=FirstX;
		Y:=FirstY;
		Radius:=Random(LightMaxAccel);
		Alpha:=Random(62800)/10000;
		SX:=Round(Cos(Alpha)*Radius)+ItSX;
		SY:=Round(Sin(Alpha)*Radius)-LightVertAdd;
		{SX:=SX+ItSX;}
		Color:=SimilarColor(ItColor);
	End
End;

Procedure InitLightLogo(FirstX,FirstY,ItColor,ItSX:LongInt);
Var
	Index:LongInt;
	Radius:LongInt;
	Alpha:Real;
	Pt:LongInt;
Begin
	For Pt:=1 To 100 Do
	Begin
		Index:=FirstFreeLight;
		If Index=0 Then Exit;
		Dec(FreeLights);
		With Light[Index] Do
		Begin
			Live:=True;
			X:=FirstX;
			Y:=FirstY;
			Radius:=A[Pt].Radius{+Random(2)};
			Alpha:=A[Pt].Alpha;
			SX:=Round(Cos(Alpha)*Radius)+ItSX;
			SY:=Round(Sin(Alpha)*Radius)-LightVertAdd;
			Color:=SimilarColor(ItColor)
		End
	End;
End;

Procedure BlowShell(Index:LongInt);
Var
	V,Num,TX,TY,TC,TS:LongInt;
Begin
	Gfx.SetPalColor(0,64,80,144);
	EraseShell(Shell[Index].X Div Scale,Shell[Index].Y Div Scale);
	Num:=MinLightsPerShell+Random(MaxLightsPerShell-MinLightsPerShell);
	TX:=Shell[Index].X;
	TY:=Shell[Index].Y;
	TC:=Shell[Index].Color;
	TS:=Shell[Index].SX;
	If (FreeLights>100) And (Random(5)=3)
		Then InitLightLogo(TX,TY,TC,TS)
		Else For V:=1 To Num Do
			InitLight(TX,TY,TC,TS);
	InitShell(Index);
	Gfx.SetPalColor(0,0,0,0)
End;

Procedure DisappearLight(Index:LongInt);
Begin
	Light[Index].Live:=False;
	Inc(FreeLights)
End;

Procedure MoveShell(Index:LongInt);
Begin
	With Shell[Index] Do
	Begin
		EraseShell(X Div Scale,Y Div Scale);
		Inc(X,SX);
		Inc(Y,SY);
		Inc(SY);
		If SY>=BlowAccel
			Then BlowShell(Index)
			Else DrawShell(X Div Scale,Y Div Scale,Color)
	End
End;

Procedure MoveLight(Index:LongInt);
Begin
	With Light[Index] Do
	Begin
		EraseLight(X Div Scale,Y Div Scale);
		Inc(X,SX);
		Inc(Y,SY);
		Inc(SY);
		If (X<Scale) Or (Y<Scale) Or
			(X>Gfx.ScrX*Scale) Or (Y>Gfx.ScrY*Scale)
			Then DisappearLight(Index)
			Else DrawLight(X Div Scale,Y Div Scale,Color)
	End
End;

Procedure Salute_OnInit;
Begin
	Count:=0;
	Gfx.Bar(0,0,Gfx.ScrX,Gfx.ScrY,clBlack);

	StartX:=Gfx.ScrX Div 2;
	StartY:=Gfx.ScrY-3;
	ShellMaxAccel:=Round(Sqrt(Gfx.ScrY))*9;
	ShellMinAccel:=ShellMaxAccel*2 Div 3;
	ShellMaxHorizAccel:=Gfx.ScrX Div 14;
	LightMaxAccel:=Gfx.ScrY Div 12;
	LightVertAdd:=Gfx.ScrY Div 10;
	MaxLightsPerShell:=120{-Gfx.ScrY Div 16};
	FreeLights:=LightsNum;

	For I:=1 To LightsNum Do
		Light[I].Live:=False;
	For I:=1 To ShellsNum Do
		InitShell(I)
End;


Procedure Salute_OnIdle;
Begin
	Inc(Count);
	If Count<(D.MachineSpeed Div 15) Then Exit;
	Count:=0;
	For I:=1 To ShellsNum Do
		MoveShell(I);
	For I:=1 To LightsNum Do
		If Light[I].Live
			Then MoveLight(I)
			Else EraseLight(1,1)
End;

Begin
	If D.FileExist('pntlogo.dat') Then
	Begin
		Assign(F,'pntlogo.dat');
		Reset(F);
		For I:=1 To 100 Do
			Read(F,A[I]);
		Close(F)
	End
	Else Begin
		Log.ErrMsg('Salute','(init)','File not found: PNTLOGO.DAT');
		For I:=1 To 100 Do
		Begin
			A[I].Alpha:=I/25*Pi;
			A[I].Radius:=15*(((I-1) Div 50)+1)
		End
	End;

	Kernel.RegScrSaver('Salute',Salute_OnInit,Salute_OnIdle)
End.