{$A+,B-,D+,E-,F-,G+,I-,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}
{
Unit Version 1.2
Last change : 10\15\2000
}

unit Stream32;

interface

uses
	Objects,NewFront;

const

{ TStream access modes }

  stCreate    = $3C00;           { Create new file }
  stOpenRead  = $3D00;           { Read access only }
  stOpenWrite = $3D01;           { Write access only }
  stOpen      = $3D02;           { Read and write access }

{ TStream error codes }

  stOk         =  0;              { No error }
  stError      = -1;              { Access error }
  stInitError  = -2;              { Cannot initialize stream }
  stReadError  = -3;              { Read beyond end of stream }
  stWriteError = -4;              { Cannot expand stream }
  stGetError   = -5;              { Get of unregistered object type }
  stPutError   = -6;              { Put of unregistered object type }

type
	PStream32=^TStream32;
	TStream32=object(TStream)
		procedure Read(var Buf; Count: Word); virtual;
		procedure Write(var Buf; Count: Word); virtual;
		function Read32(Buf: Pointer48; Count: Longint) : longint; virtual;
		function Write32(Buf: Pointer48; Count: Longint) : longint; virtual;
                procedure TestStream32; virtual;
	end;
	PFileStream32=^TFileStream32;
	TFileStream32=object(TStream32)
	private
		Data: file;
	public
		constructor Init(FName: FNameStr; Mode: Word);
		function Read32(Buf: Pointer48; Count: Longint) : longint; virtual;
		function Write32(Buf: Pointer48; Count: Longint) : longint; virtual;
		function GetPos: Longint; virtual;
		function GetSize: Longint; virtual;
		procedure Seek(Pos: Longint); virtual;
		procedure Truncate; virtual;
		destructor Done; virtual;
	end;
	PMemStream32=^TMemStream32;
	TMemStream32=object(TStream32)
	private
		Data: Pointer48;
		Size,Position: Longint;
		Created: Boolean;
	public
		constructor Init(AData: Pointer48; ASize: Longint);
		function Read32(Buf: Pointer48; Count: Longint) : longint; virtual;
		function Write32(Buf: Pointer48; Count: Longint) : longint; virtual;
		function GetPos: Longint; virtual;
		function GetSize: Longint; virtual;
		procedure Seek(Pos: Longint); virtual;
		procedure Truncate; virtual;
		destructor Done; virtual;
	end;
        PSubStream32=^TSubStream32;
	TSubStream32=object(TStream32)
	private
		Data: PStream;
	public
		constructor Init(AData: PStream);
		procedure Assign(AData: PStream);
		function Read32(Buf: Pointer48; Count: Longint) : LongInt; virtual;
		function Write32(Buf: Pointer48; Count: Longint) : LongInt; virtual;
		function GetPos: Longint; virtual;
		function GetSize: Longint; virtual;
		procedure Seek(Pos: Longint); virtual;
		procedure Truncate; virtual;
	end;

function IsStream32(AStream: PStream): Boolean;
function AsStream32(AStream: PStream): PStream32;


implementation
var
  SubStream32 : TSubStream32;

procedure TStream32.Read(var Buf; Count: Word);
begin
	Read32(farptr(@Buf),Count);
end;

procedure TStream32.Write(var Buf; Count: Word);
begin
	Write32(farptr(@Buf),Count);
end;

function TStream32.Read32(Buf: Pointer48; Count: Longint) : longint;
begin
	Abstract;
end;

function TStream32.Write32(Buf: Pointer48; Count: Longint) : longint;
begin
	Abstract;
end;

procedure TStream32.TestStream32;
begin
  Abstract;
end;

constructor TFileStream32.Init(FName: FNameStr; Mode: Word);
var
	OldMode: Word;
	Result: Integer;
begin
	inherited Init;
	OldMode:=FileMode;
	Assign(Data,FName);
	InOutRes:=0;
	if (Mode and $FF00) = stCreate then begin
		FileMode:=2 or Mode and $FF;
		Rewrite(Data,1);
	end else begin
		FileMode:=Mode and $FF;
		System.Reset(Data,1);
	end;
	FileMode:=OldMode;
	Result:=IOResult;
{  filemode := 2;}
	if Result<>0 then
		Error(stInitError,Result);
end;

function TFileStream32.Read32(Buf: Pointer48; Count: Longint) : longint;
var
	RealCount: Longint;
	Result: Integer;
begin
	if Status=stOk then begin
		InOutRes:=0;
		BlockReadWrite32(Data,Buf,Count,RealCount,False);
		Result:=IOResult;
		if (Result<>0) or (Count<>RealCount) then
			Error(stReadError,Result);
    read32 := realcount;
	end else read32 := 0;
end;

function TFileStream32.Write32(Buf: Pointer48; Count: Longint) : longint;
var
	RealCount: Longint;
	Result: Integer;
begin
	if Status=stOk then begin
		InOutRes:=0;
		BlockReadWrite32(Data,Buf,Count,RealCount,True);
		Result:=IOResult;
		if (Result<>0) or (Count<>RealCount) then
			Error(stWriteError,Result);
    write32 := realcount;
	end else write32 := 0;
end;

function TFileStream32.GetPos: Longint;
begin
	GetPos:=FilePos(Data);
end;

function TFileStream32.GetSize: Longint;
begin
	GetSize:=FileSize(Data);
end;

procedure TFileStream32.Seek(Pos: Longint);
var
	Result: Integer;
begin
	if Status=stOk then begin
		InOutRes:=0;
		System.Seek(Data,Pos);
		Result:=IOResult;
		if Result<>0 then
			Error(stError,Result);
	end;
end;

procedure TFileStream32.Truncate;
var
	Result: Integer;
begin
	if Status=stOk then begin
		InOutRes:=0;
		System.Truncate(Data);
		Result:=IOResult;
		if Result<>0 then
			Error(stError,Result);
	end;
end;

destructor TFileStream32.Done;
var
	Result: Integer;
begin
	if Status<>stInitError then begin
		InOutRes:=0;
		Close(Data);
		Result:=IOResult;
		if Result<>0 then
			Error(stError,Result);
	end;
	inherited Done;
end;

constructor TMemStream32.Init(AData: Pointer48; ASize: Longint);
var
	Selector: TSelector;
begin
	inherited Init;
	Size:=ASize;
	Created:=not Assigned32(AData) and (Size>0);
	if Created then begin
		GetMem32(Selector,ASize);
		Data:=Ptr48(Selector,0);
	end else
		Data:=AData;
end;

function TMemStream32.Read32(Buf: Pointer48; Count: Longint) : longint;
var
	OverEnd: Boolean;
begin
  read32 := 0;
	if Status=stOk then begin
		OverEnd:=Count>Size-Position;
		if OverEnd and (Count>0) then
			Count:=Size-Position;
		if Count>0 then begin
			Move32(Ptr48(Seg48(Data),Ofs48(Data)+Position),Buf,Count);
			Inc(Position,Count);
      read32 := count;
		end;
		if OverEnd or (Count<0) then
			Error(stReadError,0);
	end;
end;

function TMemStream32.Write32(Buf: Pointer48; Count: Longint) : longint;
var
	OverEnd: Boolean;
begin
  write32 := 0;
	if Status=stOk then begin
		OverEnd:=Count>Size-Position;
		if OverEnd and (Count>0) then
			Count:=Size-Position;
		if Count>0 then begin
			Move32(Buf,Ptr48(Seg48(Data),Ofs48(Data)+Position),Count);
			Inc(Position,Count);
      write32 := count;
		end;
		if OverEnd or (Count<0) then
			Error(stReadError,0);
	end;
end;

function TMemStream32.GetPos: Longint;
begin
	GetPos:=Position;
end;

function TMemStream32.GetSize: Longint;
begin
	GetSize:=Size;
end;

procedure TMemStream32.Seek(Pos: Longint);
begin
	if Status=stOk then
		if (Pos<0) or (Pos>Size) then
			Error(stError,0)
		else
			Position:=Pos;
end;

procedure TMemStream32.Truncate;
begin
	Error(stError,0)
end;

destructor TMemStream32.Done;
var
	Selector: TSelector;
begin
	if Created then begin
		Selector:=Seg48(Data);
		FreeMem32(Selector);
	end;
	inherited Done;
end;

constructor TSubStream32.Init(AData: PStream);
begin
	inherited Init;
	Assign(AData);
end;

procedure TSubStream32.Assign(AData: PStream);
begin
	Data:=AData;
	if Assigned(Data) then
		with Data^ do begin
			self.Status:=Status;
			self.ErrorInfo:=ErrorInfo;
		end;
end;

function TSubStream32.Read32(Buf: Pointer48; Count: Longint) : LongInt;
var
  P : Pointer;
  HasRead : LongInt;
  ToRead : Word;
begin
  Read32:=0;
	if Assigned(Data) then begin
                P:=NearPtr(Buf);
                HasRead:=0;
                ToRead:=$8000;
                repeat
                  if ToRead>Count-HasRead then
                    ToRead:=Count-HasRead;
		  Data^.Read(P^,ToRead);
                  asm
                    add word ptr P,8000h
                    jno @NoSelectorChange
                    mov ax,SelectorInc
                    add word ptr P+2,ax
                    @NoSelectorChange:
                  end;
                  Inc(HasRead,ToRead);
                until (HasRead=Count)or(Data^.Status<>stOk);
                Read32:=HasRead;
		with Data^ do begin
			self.Status:=Status;
			self.ErrorInfo:=ErrorInfo;
		end;
	end else
		Error(stError,0);
end;

function TSubStream32.Write32(Buf: Pointer48; Count: Longint) : LongInt;
var
  P : Pointer;
  HasWrote : LongInt;
  ToWrite : Word;
begin
  Write32:=0;
	if Assigned(Data) then begin
                P:=NearPtr(Buf);
                HasWrote:=0;
                ToWrite:=$8000;
                repeat
                  if ToWrite>Count-HasWrote then
                    ToWrite:=Count-HasWrote;
		  Data^.Write(P^,ToWrite);
                  asm
                    add word ptr P,8000h
                    jno @NoSelectorChange
                    mov ax,SelectorInc
                    add word ptr P+2,ax
                    @NoSelectorChange:
                  end;
                  Inc(HasWrote,ToWrite);
                until (HasWrote=Count)or(Data^.Status<>stOk);
                Write32:=HasWrote;
		with Data^ do begin
			self.Status:=Status;
			self.ErrorInfo:=ErrorInfo;
		end;
	end else
		Error(stError,0);
end;

function TSubStream32.GetPos: Longint;
begin
	if Assigned(Data) then begin
		GetPos:=Data^.GetPos;
		with Data^ do begin
			self.Status:=Status;
			self.ErrorInfo:=ErrorInfo;
		end;
	end else
		Error(stError,0);
end;

function TSubStream32.GetSize: Longint;
begin
	if Assigned(Data) then begin
		GetSize:=Data^.GetSize;
		with Data^ do begin
			self.Status:=Status;
			self.ErrorInfo:=ErrorInfo;
		end;
	end else
		Error(stError,0);
end;

procedure TSubStream32.Seek(Pos: Longint);
begin
	if Assigned(Data) then begin
		Data^.Seek(Pos);
		with Data^ do begin
			self.Status:=Status;
			self.ErrorInfo:=ErrorInfo;
		end;
	end else
		Error(stError,0);
end;

procedure TSubStream32.Truncate;
begin
	if Assigned(Data) then begin
		Data^.Truncate;
		with Data^ do begin
			self.Status:=Status;
			self.ErrorInfo:=ErrorInfo;
		end;
	end else
		Error(stError,0);
end;

function IsStream32(AStream: PStream): Boolean; assembler;
const
	MyVMT: Word=Ofs(TypeOf(TStream32)^);
type
	VMT=record
		Size,NegSize,DMT,Reserved: Word;
		Done,Error,Flush,GetPos,GetSize,Read,Seek,Truncate,
                Write,Read32,Write32,TestStream32: Pointer;
	end;
asm
	les di,[AStream]
	mov di,es:[di]
	mov si,MyVMT
db $66; mov ax,word ptr ds:[si+VMT.TestStream32]
db $66; sub ax,word ptr ds:[di+VMT.TestStream32]
        db $0F,$94,$C0 { sete al }
end;

function AsStream32(AStream: PStream): PStream32;
begin
	if IsStream32(AStream) then
		AsStream32:=PStream32(AStream)
	else begin
		SubStream32.Assign(AStream);
		AsStream32:=@SubStream32;
	end;
end;


begin
  SubStream32.Init(nil);
end.