{ ************ Trackbar ************ }
Type
	TTrack=Record
		Kind: TTrackKind;
		Min, Max, Value: Integer;
		Ticks: Boolean;
		OnUpdate:TCtlHandler;
		Down: Boolean;
	End;

Var
	PicTrack: TSkinPic;
	TrackHalfWidth: Integer;


Procedure Track_DrawStick(Me:PControl; X, Y1, Y2:Integer);
Begin
	With Me^, TTrack(Data^) Do
	Begin
		{BCRT.Delay(7);}
		{Gfx.Bar(X, Place.Y1, X+1, Place.Y2, clSilver);}
		If Enabled Then
		Begin
			Gfx.Line(X, Y1+1, X, Y2-1, Kernel.ClrLines);
			Gfx.PutPixel(X, Y1, Kernel.ClrLines2);
			Gfx.PutPixel(X, Y2, Kernel.ClrLines2)
		End
		Else Begin
			Gfx.Line(X, Y1+1, X, Y2-1, Kernel.ClrDisabled);
			Gfx.PutPixel(X, Y1, Kernel.ClrDisabled2);
			Gfx.PutPixel(X, Y2, Kernel.ClrDisabled2)
		End
	End
End;

Procedure Track_ShowSlider(Me:PControl; Show:Boolean);
Var
	V1, CurX: Integer;
	Delta: Real;
Begin
	With Me^, Place, TTrack(Data^) Do
	Begin
		V1:=((X2-X1) Div 2)-TrackHalfWidth; { ii 箪 }
		Delta:=V1/(Max-Min);
		CurX:=X1+TrackHalfWidth+Round((Value-Min)*Delta)*2;
		If Show
			Then Gfx.PutSprite(CurX-TrackHalfWidth, Y1, PicTrack[Enabled, Down])
			Else Kernel.WindowBar(CurX-TrackHalfWidth, Y1,
				CurX+TrackHalfWidth, Y2, False, True)
	End
End;

Procedure Track_IntDraw(Me:PControl);
Var
	CurX, CenterY, I: Integer;
	Delta, DeltaStep: Real;
	V1, V2, V3: Integer;
Begin
	If ((Not Me^.CanDraw) And (Not Kernel.DontBother)) Or Not Me^.Visible Then
		Exit;
	With Me^, Place, TTrack(Data^) Do
	Begin
		{Kernel.WindowBar(X1, Y1, X2, Y2, False, True);}
		Case Kind Of
			tkMonowidth:Begin
				CurX:=X1;
				For I:=0 To (X2-X1) Div 2 Do
				Begin
					Track_DrawStick(Me, CurX, Y1+4, Y2-4);
					Inc(CurX, 2)
				End;
			End;
			tkGrowRight:Begin
				CurX:=X1;
				CenterY:=Y1+(Y2-Y1) Div 2;
				Delta:=2;
				DeltaStep:=6/((X2-X1) Div 2);
				For I:=0 To (X2-X1) Div 2 Do
				Begin
					Track_DrawStick(Me, CurX, CenterY-Round(Delta), CenterY+Round(Delta));
					Inc(CurX, 2);
					Delta:=Delta+DeltaStep
				End
			End
		End;

		V1:=((X2-X1) Div 2)-TrackHalfWidth-1; { ii 箪 }
		Delta:=V1/(Max-Min);
		If Ticks Then
			For I:=0 To Max-Min Do
			Begin
				CurX:=X1+TrackHalfWidth+1+Round(I*Delta)*2;
				Track_DrawStick(Me, CurX, Y1, Y2)
			End;
		Track_ShowSlider(Me, True)
	End;
End;

Procedure Track_Paint(Me:PControl; VX1, VY1, VX2:Integer);
Begin
	If Odd(VX2-VX1) Then
		Dec(VX2);
	With Me^.Owner^.ClientRect Do
		D.AssignRect(X1+VX1, Y1+VY1, X1+VX2, Y1+VY1+20, Me^.Place);
	Track_IntDraw(Me)
End;

Procedure Track_Draw(Me:PControl);
Begin
	With Me^ Do
		If CanDraw Or Kernel.DontBother Then
		Begin
			Kernel.BeginPaint(Place);
			If Visible
				Then Track_IntDraw(Me)
				Else Kernel.HideCtl(Me);
			Kernel.EndPaint
		End
End;


Procedure Track_NewValue(Me:PControl; NewValue:Integer);
Begin
	With Me^, TTrack(Data^) Do
	Begin
		Track_ShowSlider(Me, False);
		Value:=NewValue;
		Track_IntDraw(Me);
		OnUpdate(Me^.Owner, Me)
	End
End;

Function Track_CoordToValue(Me:PControl; X:Integer): Integer;
Begin
	With Me^, Place, TTrack(Data^) Do
	Begin
		D.ToRange(X, X1+TrackHalfWidth, X2-TrackHalfWidth);
		Track_CoordToValue:=Round(LongInt(X-X1-TrackHalfWidth)*
			LongInt(Max-Min)/LongInt(X2-X1-TrackHalfWidth*2)+Min)
	End
End;

Procedure Track_ShowMouse(Var X, Y:Integer);
Var
	TempValue: Integer;
Begin
	With CurCtl^, Place, TTrack(Data^) Do
	Begin
		D.ToRange(X, X1, X2);
		Y:=LastMsY;
		TempValue:=Track_CoordToValue(CurCtl, X);
		If TempValue<>Value Then
			Track_NewValue(CurCtl, TempValue)
	End;
	Kernel.ShowMouse(X, Y)
End;

Procedure Track_OnClick(Me:PControl; X, Y:Integer);
Begin
	With Me^, TTrack(Data^) Do
	Begin
		{Track_NewValue(Me, Track_CoordToValue(Me, X));}
		Track_ShowSlider(Me, False);
		Value:=Track_CoordToValue(Me, X);
		OnUpdate(Me^.Owner, Me);
		Down:=True;
		Track_IntDraw(Me);
		CurCtl:=Me;
		LastMsY:=Y;

		Repeat
			Kernel.CheckSysKeys
		Until Kernel.HeartOfPolit(Kernel.HideMouse, Track_ShowMouse,
			Kernel.RefreshNeeded)=0;
		Down:=False;
		Track_ShowSlider(Me, True)
	End
End;

Function Track_OnKey(Me:PControl; Key, ExtKey:Char): Boolean;
Var
	Recognized: Boolean;
	NewValue: Integer;
Begin
	Recognized:=False;
	With Me^, TTrack(Data^) Do
	Begin
		If Key=#0 Then
			Case ExtKey Of
				#75, #72:Begin { left, up }
					NewValue:=D.Max(Min, Value-1);
					Recognized:=True
				End;
				#77, #80:Begin { right, down }
					NewValue:=D.Min(Max, Value+1);
					Recognized:=True
				End;
				#71:Begin { home }
					NewValue:=Min;
					Recognized:=True
				End;
				#79:Begin { end }
					NewValue:=Max;
					Recognized:=True
				End;
			End;

		If Recognized Then
			If NewValue<>Value Then
				Track_NewValue(Me, NewValue)
	End;
	Track_OnKey:=Recognized
End;

Function Track_GetPos(Me:PControl): Integer;
Begin
	Track_GetPos:=TTrack(Me^.Data^).Value
End;

Procedure Track_SetPos(Me:PControl; VPos:Integer);
Begin
	With Me^, TTrack(Data^) Do
		If AutoRedraw And Not Kernel.IgnoreAutoRedraw And
			(Me^.CanDraw Or Kernel.DontBother) And Visible Then
		Begin
			Kernel.BeginPaint(Place);
			Track_ShowSlider(Me, False);
			Value:=VPos;
			D.ToRange(Value, Min, Max);
			Track_IntDraw(Me);
			Kernel.EndPaint
		End
		Else Begin
			Value:=VPos;
			D.ToRange(Value, Min, Max)
		End
End;

Procedure Track_OnDestroy(Me:PControl);
Begin
	With Me^ Do
		FreeMem(Data, SizeOf(TTrack));
End;

Function Track_Create(Prog:PProg; VMin, VMax, VValue:Integer;
	VKind:TTrackKind; VTicks:Boolean; VOnUpdate:TCtlHandler): PControl;
Var
	Me:PControl;
Begin
	Me:=Kernel.AddCtl(Prog);
		{Track_OnClick, Track_OnKey, Kernel.CtlRefocusDummy,
		Kernel.CtlDummy, Track_OnDestroy, VEnabled, True, Hint);}
	With Me^ Do
	Begin
		OnClick:=Track_OnClick;
		OnKey:=Track_OnKey;
		OnDestroy:=Track_OnDestroy;

		GetMem(Data, SizeOf(TTrack));
		With TTrack(Data^) Do
		Begin
			Kind:=VKind;
			Min:=VMin;
			Max:=VMax;
			Value:=VValue;
			Ticks:=VTicks;
			OnUpdate:=VOnUpdate;
			Down:=False;

		End
	End;
	Track_Create:=Me
End;
